#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
This script is used for Validation of the new numba version of the old 'beamformer.cpp'.
This is the pre-acoular integration check.

Essentially the NUMBA versions are checked againts the old WEAVE versions AND 
against NUMPY code (which is much clearer to varify with the bare eye).

This script needs the 'fastFuncs.py' with all the NUMBA optimized code and the 
'beamformer.so' with all the WEAVE code in its directory.
--> Only runs under python=2

Created on Tue Aug 22 12:41:50 2017

@author: tomgensch
"""
import numpy as np
import time as tm

import beamformer  # the old weave benchmark
import fastFuncs as beamNew  # The new module

def vectorized(csm, e, h, r0, rm, kj, normalizeFactor):
    nFreqs = csm.shape[0]
    nGridPoints = r0.shape[0]
    beamformOutput = np.zeros((nFreqs, nGridPoints), np.complex128)
    for cntFreqs in xrange(nFreqs):
        for cntGrid in xrange(nGridPoints):
            steeringVector = np.exp(-1j * kj[cntFreqs].imag * (rm[cntGrid, :] - r0[cntGrid]))
            beamformOutput[cntFreqs, cntGrid] = np.inner(np.inner(csm[cntFreqs, :, :], steeringVector), steeringVector.conj())
    return beamformOutput.real / normalizeFactor

def beamformerRef(csm, steer):
    nFreqs = csm.shape[0]
    nGridPoints = r0.shape[0]
    beamformOutput = np.zeros((nFreqs, nGridPoints), np.complex128)
    for cntFreqs in xrange(nFreqs):
        for cntGrid in xrange(nGridPoints):
            beamformOutput[cntFreqs, cntGrid] = np.inner(np.inner(csm[cntFreqs, :, :], steer[cntFreqs, cntGrid, :]), steer[cntFreqs, cntGrid, :].conj())
    return beamformOutput.real


nFreqs = 1
nMics = 56
nGridPoints = 100#1521

csm = np.random.rand(nFreqs, nMics, nMics) + 1j*np.random.rand(nFreqs, nMics, nMics)  # cross spectral matrix
for cntFreqs in range(nFreqs):
    csm[cntFreqs, :, :] += csm[cntFreqs, :, :].T.conj()  # make CSM hermetical
csmRemovedDiag = np.array(csm)
[np.fill_diagonal(csmRemovedDiag[cntFreq, :, :], 0.0) for cntFreq in range(csmRemovedDiag.shape[0])]

e = np.random.rand(nMics) + 1j*np.random.rand(nMics)  # has no usage
h = np.zeros((nFreqs, nGridPoints))  # results are stored here, if function has no return value
r0 = np.random.rand(nGridPoints)  # distance between gridpoints and middle of array
rm = np.random.rand(nGridPoints, nMics)  # distance between gridpoints and all mics in the array
kj = np.zeros(nFreqs) + 1j*np.random.rand(nFreqs) # complex
eigVal, eigVec = np.linalg.eigh(csm)
eigValTrans, eigVecTrans = np.linalg.eigh(csm.transpose(0,2,1))  # of transpose CSM, to compare with old weave mistake
indLow = 0
indHigh = nMics

#%% Validate the PSF-formulations - Against Numpy
steerFormulation1 = np.zeros((nFreqs, nGridPoints, nMics), np.complex128)
steerFormulation2 = np.zeros((nFreqs, nGridPoints, nMics), np.complex128)
steerFormulation3 = np.zeros((nFreqs, nGridPoints, nMics), np.complex128)
steerFormulation4 = np.zeros((nFreqs, nGridPoints, nMics), np.complex128)
steerFormulation1normalized = np.zeros((nFreqs, nGridPoints, nMics), np.complex128)
steerFormulation2normalized = np.zeros((nFreqs, nGridPoints, nMics), np.complex128)
steerFormulation3normalized = np.zeros((nFreqs, nGridPoints, nMics), np.complex128)
steerFormulation4normalized = np.zeros((nFreqs, nGridPoints, nMics), np.complex128)
transfer = np.zeros((nFreqs, nGridPoints, nMics), np.complex128)
csmPsf = np.zeros((nFreqs, nMics, nMics), np.complex128)

indSource = [1, 3]
indSourceCalcWithNumpy = 0

for cntFreqs in xrange(nFreqs):
    for cntGrid in xrange(nGridPoints):
        steerFormulation1[cntFreqs, cntGrid, :] = np.exp(-1j * kj[cntFreqs].imag * (rm[cntGrid, :] - r0[cntGrid])) / nMics
        steerFormulation2[cntFreqs, cntGrid, :] = np.exp(-1j * kj[cntFreqs].imag * (rm[cntGrid, :] - r0[cntGrid])) * rm[cntGrid, :] / r0[cntGrid] / nMics
        steerFormulation3[cntFreqs, cntGrid, :] = np.exp(-1j * kj[cntFreqs].imag * (rm[cntGrid, :] - r0[cntGrid])) / rm[cntGrid, :] / r0[cntGrid] / np.sum(1.0 / rm[cntGrid, :]**2)
        steerFormulation4[cntFreqs, cntGrid, :] = np.exp(-1j * kj[cntFreqs].imag * (rm[cntGrid, :] - r0[cntGrid])) / rm[cntGrid, :] / np.sqrt(nMics * np.sum(1.0 / rm[cntGrid, :]**2))
        
        steerFormulation1normalized[cntFreqs, cntGrid, :] = steerFormulation1[cntFreqs, cntGrid, :] / np.sqrt(np.vdot(steerFormulation1[cntFreqs, cntGrid, :], steerFormulation1[cntFreqs, cntGrid, :]))
        steerFormulation2normalized[cntFreqs, cntGrid, :] = steerFormulation2[cntFreqs, cntGrid, :] / np.sqrt(np.vdot(steerFormulation2[cntFreqs, cntGrid, :], steerFormulation2[cntFreqs, cntGrid, :]))
        steerFormulation3normalized[cntFreqs, cntGrid, :] = steerFormulation3[cntFreqs, cntGrid, :] / np.sqrt(np.vdot(steerFormulation3[cntFreqs, cntGrid, :], steerFormulation3[cntFreqs, cntGrid, :]))
        steerFormulation4normalized[cntFreqs, cntGrid, :] = steerFormulation4[cntFreqs, cntGrid, :] / np.sqrt(np.vdot(steerFormulation4[cntFreqs, cntGrid, :], steerFormulation4[cntFreqs, cntGrid, :]))
        
        transfer[cntFreqs, cntGrid, :] = np.exp(-1j * kj[cntFreqs].imag * (rm[cntGrid, :] - r0[cntGrid])) * r0[cntGrid] / rm[cntGrid, :]
    csmPsf[cntFreqs, :] = np.outer(transfer[cntFreqs, indSource[indSourceCalcWithNumpy], :], transfer[cntFreqs, indSource[indSourceCalcWithNumpy], :].conj())

psf1 = beamNew.calcPointSpreadFunction(1, (r0, rm, kj, indSource))[:, :, indSourceCalcWithNumpy]
psf1Ref = beamformerRef(csmPsf, steerFormulation1)
relDiff = (psf1 - psf1Ref) / (psf1 + psf1Ref) * 2
psf1DiffNumpy = np.amax(np.amax(relDiff, 1), 0)

psf2 = beamNew.calcPointSpreadFunction(2, (r0, rm, kj, indSource))[:, :, indSourceCalcWithNumpy]
psf2Ref = beamformerRef(csmPsf, steerFormulation2)
relDiff = (psf2 - psf2Ref) / (psf2 + psf2Ref) * 2
psf2DiffNumpy = np.amax(np.amax(relDiff, 1), 0)

psf3 = beamNew.calcPointSpreadFunction(3, (r0, rm, kj, indSource))[:, :, indSourceCalcWithNumpy]
psf3Ref = beamformerRef(csmPsf, steerFormulation3)
relDiff = (psf3 - psf3Ref) / (psf3 + psf3Ref) * 2
psf3DiffNumpy = np.amax(np.amax(relDiff, 1), 0)

psf4 = beamNew.calcPointSpreadFunction(4, (r0, rm, kj, indSource))[:, :, indSourceCalcWithNumpy]
psf4Ref = beamformerRef(csmPsf, steerFormulation4)
relDiff = (psf4 - psf4Ref) / (psf4 + psf4Ref) * 2
psf4DiffNumpy = np.amax(np.amax(relDiff, 1), 0)

### --> looks good, except for formulation 4, where there is a known mistake (see Issue #5 in gitlab)

#%% Validate the PSF-formulations - Against Weave

indSource = [0, 42, 80]
hPSF = np.zeros((nGridPoints, len(indSource)), np.float64)

psf1 = beamNew.calcPointSpreadFunction(1, (r0, rm, kj, indSource))
beamformer.r_beam_psf1(hPSF, r0, r0[indSource], rm, rm[indSource, :], kj[0])
relDiff = (psf1[0,:,:] - hPSF) / (psf1[0,:,:] + hPSF) * 2
psf1DiffWeave = np.amax(np.amax(relDiff, 0), 0)

psf2 = beamNew.calcPointSpreadFunction(2, (r0, rm, kj, indSource))
beamformer.r_beam_psf2(hPSF, r0, r0[indSource], rm, rm[indSource, :], kj[0])
relDiff = (psf2[0,:,:] - hPSF) / (psf2[0,:,:] + hPSF) * 2
psf2DiffWeave = np.amax(np.amax(relDiff, 0), 0)

psf3 = beamNew.calcPointSpreadFunction(3, (r0, rm, kj, indSource))
beamformer.r_beam_psf3(hPSF, r0, r0[indSource], rm, rm[indSource, :], kj[0])
relDiff = (psf3[0,:,:] - hPSF) / (psf3[0,:,:] + hPSF) * 2
psf3DiffWeave = np.amax(np.amax(relDiff, 0), 0)

psf4 = beamNew.calcPointSpreadFunction(4, (r0, rm, kj, indSource))
beamformer.r_beam_psf4(hPSF, r0, r0[indSource], rm, rm[indSource, :], kj[0])
relDiff = (psf4[0,:,:] - hPSF) / (psf4[0,:,:] + hPSF) * 2
psf4DiffWeave = np.amax(np.amax(relDiff, 0), 0)

### --> looks good, but this means that at the moment there is a mistake in formulation 4 (see comment in last section)

#%% PSF - runtime comparison between weave and numba
t0 = tm.time()
psf1 = beamNew.calcPointSpreadFunction(1, (r0, rm, kj, indSource))
tNumba = tm.time() - t0

t0 = tm.time()
for cntFreqs in xrange(len(kj)):
    beamformer.r_beam_psf1(hPSF, r0, r0[indSource], rm, rm[indSource, :], kj[cntFreqs])
tWeave = tm.time() - t0

print('tNumba: %s' %tNumba)
print('tWeave: %s' %tWeave)

#%% new beamformer - against numpy
indLow = 0
indHigh = nMics  # when all eigenvalues are taken into account --> eigVal beamformer and conventional beamformer should give same reults

normFull = 1.0
normDiag = np.float64(nMics) / np.float64(nMics - 1)


comp1Voll, dummy = beamNew.beamformerFreq(False, 1, False, normFull, (r0, rm, kj, csm))
comp1VollRefNumpy = beamformerRef(csm, steerFormulation1)
relDiff = (comp1Voll - comp1VollRefNumpy) / (comp1Voll + comp1VollRefNumpy) * 2
error1VollAgainstNumpy = np.amax(np.amax(abs(relDiff), 0), 0)

comp1Diag, dummy = beamNew.beamformerFreq(False, 1, True, normDiag, (r0, rm, kj, csm))
comp1DiagRefNumpy = beamformerRef(csmRemovedDiag, steerFormulation1) * normDiag
relDiff = (comp1Diag - comp1DiagRefNumpy) / (comp1Diag + comp1DiagRefNumpy) * 2
error1DiagAgainstNumpy = np.amax(np.amax(abs(relDiff), 0), 0)

comp1VollEig, dummy = beamNew.beamformerFreq(True, 1, False, normFull, (r0, rm, kj, eigVal[:, indLow : indHigh], eigVec[:, :, indLow : indHigh]))
comp1VollEigRefNumpy = beamformerRef(csm, steerFormulation1)
relDiff = (comp1VollEig - comp1VollEigRefNumpy) / (comp1VollEig + comp1VollEigRefNumpy) * 2
error1VollEigAgainstNumpy = np.amax(np.amax(abs(relDiff), 0), 0)

comp1DiagEig, dummy = beamNew.beamformerFreq(True, 1, True, normDiag, (r0, rm, kj, eigVal[:, indLow : indHigh], eigVec[:, :, indLow : indHigh]))
comp1DiagEigRefNumpy = beamformerRef(csmRemovedDiag, steerFormulation1) * normDiag
relDiff = (comp1DiagEig - comp1DiagEigRefNumpy) / (comp1DiagEig + comp1DiagEigRefNumpy) * 2
error1DiagEigAgainstNumpy = np.amax(np.amax(abs(relDiff), 0), 0)



comp2Voll, dummy = beamNew.beamformerFreq(False, 2, False, normFull, (r0, rm, kj, csm))
comp2VollRefNumpy = beamformerRef(csm, steerFormulation2)
relDiff = (comp2Voll - comp2VollRefNumpy) / (comp2Voll + comp2VollRefNumpy) * 2
error2VollAgainstNumpy = np.amax(np.amax(abs(relDiff), 0), 0)

comp2Diag, dummy = beamNew.beamformerFreq(False, 2, True, normDiag, (r0, rm, kj, csm))
comp2DiagRefNumpy = beamformerRef(csmRemovedDiag, steerFormulation2) * normDiag
relDiff = (comp2Diag - comp2DiagRefNumpy) / (comp2Diag + comp2DiagRefNumpy) * 2
error2DiagAgainstNumpy = np.amax(np.amax(abs(relDiff), 0), 0)

comp2VollEig, dummy = beamNew.beamformerFreq(True, 2, False, normFull, (r0, rm, kj, eigVal[:, indLow : indHigh], eigVec[:, :, indLow : indHigh]))
comp2VollEigRefNumpy = beamformerRef(csm, steerFormulation2)
relDiff = (comp2VollEig - comp2VollEigRefNumpy) / (comp2VollEig + comp2VollEigRefNumpy) * 2
error2VollEigAgainstNumpy = np.amax(np.amax(abs(relDiff), 0), 0)

comp2DiagEig, dummy = beamNew.beamformerFreq(True, 2, True, normDiag, (r0, rm, kj, eigVal[:, indLow : indHigh], eigVec[:, :, indLow : indHigh]))
comp2DiagEigRefNumpy = beamformerRef(csmRemovedDiag, steerFormulation2) * normDiag
relDiff = (comp2DiagEig - comp2DiagEigRefNumpy) / (comp2DiagEig + comp2DiagEigRefNumpy) * 2
error2DiagEigAgainstNumpy = np.amax(np.amax(abs(relDiff), 0), 0)



comp3Voll, dummy = beamNew.beamformerFreq(False, 3, False, normFull, (r0, rm, kj, csm))
comp3VollRefNumpy = beamformerRef(csm, steerFormulation3)
relDiff = (comp3Voll - comp3VollRefNumpy) / (comp3Voll + comp3VollRefNumpy) * 2
error3VollAgainstNumpy = np.amax(np.amax(abs(relDiff), 0), 0)

comp3Diag, dummy = beamNew.beamformerFreq(False, 3, True, normDiag, (r0, rm, kj, csm))
comp3DiagRefNumpy = beamformerRef(csmRemovedDiag, steerFormulation3) * normDiag
relDiff = (comp3Diag - comp3DiagRefNumpy) / (comp3Diag + comp3DiagRefNumpy) * 2
error3DiagAgainstNumpy = np.amax(np.amax(abs(relDiff), 0), 0)

comp3VollEig, dummy = beamNew.beamformerFreq(True, 3, False, normFull, (r0, rm, kj, eigVal[:, indLow : indHigh], eigVec[:, :, indLow : indHigh]))
comp3VollEigRefNumpy = beamformerRef(csm, steerFormulation3)
relDiff = (comp3VollEig - comp3VollEigRefNumpy) / (comp3VollEig + comp3VollEigRefNumpy) * 2
error3VollEigAgainstNumpy = np.amax(np.amax(abs(relDiff), 0), 0)

comp3DiagEig, dummy = beamNew.beamformerFreq(True, 3, True, normDiag, (r0, rm, kj, eigVal[:, indLow : indHigh], eigVec[:, :, indLow : indHigh]))
comp3DiagEigRefNumpy = beamformerRef(csmRemovedDiag, steerFormulation3) * normDiag
relDiff = (comp3DiagEig - comp3DiagEigRefNumpy) / (comp3DiagEig + comp3DiagEigRefNumpy) * 2
error3DiagEigAgainstNumpy = np.amax(np.amax(abs(relDiff), 0), 0)



comp4Voll, dummy = beamNew.beamformerFreq(False, 4, False, normFull, (r0, rm, kj, csm))
comp4VollRefNumpy = beamformerRef(csm, steerFormulation4)
relDiff = (comp4Voll - comp4VollRefNumpy) / (comp4Voll + comp4VollRefNumpy) * 2
error4VollAgainstNumpy = np.amax(np.amax(abs(relDiff), 0), 0)

comp4Diag, dummy = beamNew.beamformerFreq(False, 4, True, normDiag, (r0, rm, kj, csm))
comp4DiagRefNumpy = beamformerRef(csmRemovedDiag, steerFormulation4) * normDiag
relDiff = (comp4Diag - comp4DiagRefNumpy) / (comp4Diag + comp4DiagRefNumpy) * 2
error4DiagAgainstNumpy = np.amax(np.amax(abs(relDiff), 0), 0)

comp4VollEig, dummy = beamNew.beamformerFreq(True, 4, False, normFull, (r0, rm, kj, eigVal[:, indLow : indHigh], eigVec[:, :, indLow : indHigh]))
comp4VollEigRefNumpy = beamformerRef(csm, steerFormulation4)
relDiff = (comp4VollEig - comp4VollEigRefNumpy) / (comp4VollEig + comp4VollEigRefNumpy) * 2
error4VollEigAgainstNumpy = np.amax(np.amax(abs(relDiff), 0), 0)

comp4DiagEig, dummy = beamNew.beamformerFreq(True, 4, True, normDiag, (r0, rm, kj, eigVal[:, indLow : indHigh], eigVec[:, :, indLow : indHigh]))
comp4DiagEigRefNumpy = beamformerRef(csmRemovedDiag, steerFormulation4) * normDiag
relDiff = (comp4DiagEig - comp4DiagEigRefNumpy) / (comp4DiagEig + comp4DiagEigRefNumpy) * 2
error4DiagEigAgainstNumpy = np.amax(np.amax(abs(relDiff), 0), 0)



compSpecificVoll, dummy = beamNew.beamformerFreq(False, 'custom', False, normFull, (steerFormulation4, csm))
compSpecificVollRefNumpy = beamformerRef(csm, steerFormulation4)
relDiff = (compSpecificVoll - compSpecificVollRefNumpy) / (compSpecificVoll + compSpecificVollRefNumpy) * 2
errorSpecificVollAgainstNumpy = np.amax(np.amax(abs(relDiff), 0), 0)

compSpecificDiag, dummy = beamNew.beamformerFreq(False, 'custom', True, normDiag, (steerFormulation4, csm))
compSpecificDiagRefNumpy = beamformerRef(csmRemovedDiag, steerFormulation4) * normDiag
relDiff = (compSpecificDiag - compSpecificDiagRefNumpy) / (compSpecificDiag + compSpecificDiagRefNumpy) * 2
errorSpecificDiagAgainstNumpy = np.amax(np.amax(abs(relDiff), 0), 0)

compSpecificVollEig, dummy = beamNew.beamformerFreq(True, 'custom', False, normFull, (steerFormulation4, eigVal[:, indLow : indHigh], eigVec[:, :, indLow : indHigh]))
compSpecificVollEigRefNumpy = beamformerRef(csm, steerFormulation4)
relDiff = (compSpecificVollEig - compSpecificVollEigRefNumpy) / (compSpecificVollEig + compSpecificVollEigRefNumpy) * 2
errorSpecificVollEigAgainstNumpy = np.amax(np.amax(abs(relDiff), 0), 0)

compSpecificDiagEig, dummy = beamNew.beamformerFreq(True, 'custom', True, normDiag, (steerFormulation4, eigVal[:, indLow : indHigh], eigVec[:, :, indLow : indHigh]))
compSpecificDiagEigRefNumpy = beamformerRef(csmRemovedDiag, steerFormulation4) * normDiag
relDiff = (compSpecificDiagEig - compSpecificDiagEigRefNumpy) / (compSpecificDiagEig + compSpecificDiagEigRefNumpy) * 2
errorSpecificDiagEigAgainstNumpy = np.amax(np.amax(abs(relDiff), 0), 0)






## now check the normalization of steering vector functionality

comp1VollNormalizedSteer, steerNorm = beamNew.beamformerFreq(False, 1, False, normFull, (r0, rm, kj, csm))
comp1VollNormalizedSteer /= steerNorm
comp1VollNormalizedSteerRefNumpy = beamformerRef(csm, steerFormulation1normalized)
relDiff = (comp1VollNormalizedSteer - comp1VollNormalizedSteerRefNumpy) / (comp1VollNormalizedSteer + comp1VollNormalizedSteerRefNumpy) * 2
error1VollNormalizedSteerAgainstNumpy = np.amax(np.amax(abs(relDiff), 0), 0)

comp1DiagNormalizedSteer, steerNorm = beamNew.beamformerFreq(False, 1, True, normDiag, (r0, rm, kj, csm))
comp1DiagNormalizedSteer /= steerNorm
comp1DiagNormalizedSteerRefNumpy = beamformerRef(csmRemovedDiag, steerFormulation1normalized) * normDiag
relDiff = (comp1DiagNormalizedSteer - comp1DiagNormalizedSteerRefNumpy) / (comp1DiagNormalizedSteer + comp1DiagNormalizedSteerRefNumpy) * 2
error1DiagNormalizedSteerAgainstNumpy = np.amax(np.amax(abs(relDiff), 0), 0)

comp1VollEigNormalizedSteer, steerNorm = beamNew.beamformerFreq(True, 1, False, normFull, (r0, rm, kj, eigVal[:, indLow : indHigh], eigVec[:, :, indLow : indHigh]))
comp1VollEigNormalizedSteer /= steerNorm
comp1VollEigNormalizedSteerRefNumpy = beamformerRef(csm, steerFormulation1normalized)
relDiff = (comp1VollEigNormalizedSteer - comp1VollEigNormalizedSteerRefNumpy) / (comp1VollEigNormalizedSteer + comp1VollEigNormalizedSteerRefNumpy) * 2
error1VollEigNormalizedSteerAgainstNumpy = np.amax(np.amax(abs(relDiff), 0), 0)

comp1DiagEigNormalizedSteer, steerNorm = beamNew.beamformerFreq(True, 1, True, normDiag, (r0, rm, kj, eigVal[:, indLow : indHigh], eigVec[:, :, indLow : indHigh]))
comp1DiagEigNormalizedSteer /= steerNorm
comp1DiagEigNormalizedSteerRefNumpy = beamformerRef(csmRemovedDiag, steerFormulation1normalized) * normDiag
relDiff = (comp1DiagEigNormalizedSteer - comp1DiagEigNormalizedSteerRefNumpy) / (comp1DiagEigNormalizedSteer + comp1DiagEigNormalizedSteerRefNumpy) * 2
error1DiagEigNormalizedSteerAgainstNumpy = np.amax(np.amax(abs(relDiff), 0), 0)



comp2VollNormalizedSteer, steerNorm = beamNew.beamformerFreq(False, 2, False, normFull, (r0, rm, kj, csm))
comp2VollNormalizedSteer /= steerNorm
comp2VollNormalizedSteerRefNumpy = beamformerRef(csm, steerFormulation2normalized)
relDiff = (comp2VollNormalizedSteer - comp2VollNormalizedSteerRefNumpy) / (comp2VollNormalizedSteer + comp2VollNormalizedSteerRefNumpy) * 2
error2VollNormalizedSteerAgainstNumpy = np.amax(np.amax(abs(relDiff), 0), 0)

comp2DiagNormalizedSteer, steerNorm = beamNew.beamformerFreq(False, 2, True, normDiag, (r0, rm, kj, csm))
comp2DiagNormalizedSteer /= steerNorm
comp2DiagNormalizedSteerRefNumpy = beamformerRef(csmRemovedDiag, steerFormulation2normalized) * normDiag
relDiff = (comp2DiagNormalizedSteer - comp2DiagNormalizedSteerRefNumpy) / (comp2DiagNormalizedSteer + comp2DiagNormalizedSteerRefNumpy) * 2
error2DiagNormalizedSteerAgainstNumpy = np.amax(np.amax(abs(relDiff), 0), 0)

comp2VollEigNormalizedSteer, steerNorm = beamNew.beamformerFreq(True, 2, False, normFull, (r0, rm, kj, eigVal[:, indLow : indHigh], eigVec[:, :, indLow : indHigh]))
comp2VollEigNormalizedSteer /= steerNorm
comp2VollEigNormalizedSteerRefNumpy = beamformerRef(csm, steerFormulation2normalized)
relDiff = (comp2VollEigNormalizedSteer - comp2VollEigNormalizedSteerRefNumpy) / (comp2VollEigNormalizedSteer + comp2VollEigNormalizedSteerRefNumpy) * 2
error2VollEigNormalizedSteerAgainstNumpy = np.amax(np.amax(abs(relDiff), 0), 0)

comp2DiagEigNormalizedSteer, steerNorm = beamNew.beamformerFreq(True, 2, True, normDiag, (r0, rm, kj, eigVal[:, indLow : indHigh], eigVec[:, :, indLow : indHigh]))
comp2DiagEigNormalizedSteer /= steerNorm
comp2DiagEigNormalizedSteerRefNumpy = beamformerRef(csmRemovedDiag, steerFormulation2normalized) * normDiag
relDiff = (comp2DiagEigNormalizedSteer - comp2DiagEigNormalizedSteerRefNumpy) / (comp2DiagEigNormalizedSteer + comp2DiagEigNormalizedSteerRefNumpy) * 2
error2DiagEigNormalizedSteerAgainstNumpy = np.amax(np.amax(abs(relDiff), 0), 0)



comp3VollNormalizedSteer, steerNorm = beamNew.beamformerFreq(False, 3, False, normFull, (r0, rm, kj, csm))
comp3VollNormalizedSteer /= steerNorm
comp3VollNormalizedSteerRefNumpy = beamformerRef(csm, steerFormulation3normalized)
relDiff = (comp3VollNormalizedSteer - comp3VollNormalizedSteerRefNumpy) / (comp3VollNormalizedSteer + comp3VollNormalizedSteerRefNumpy) * 2
error3VollNormalizedSteerAgainstNumpy = np.amax(np.amax(abs(relDiff), 0), 0)

comp3DiagNormalizedSteer, steerNorm = beamNew.beamformerFreq(False, 3, True, normDiag, (r0, rm, kj, csm))
comp3DiagNormalizedSteer /= steerNorm
comp3DiagNormalizedSteerRefNumpy = beamformerRef(csmRemovedDiag, steerFormulation3normalized) * normDiag
relDiff = (comp3DiagNormalizedSteer - comp3DiagNormalizedSteerRefNumpy) / (comp3DiagNormalizedSteer + comp3DiagNormalizedSteerRefNumpy) * 2
error3DiagNormalizedSteerAgainstNumpy = np.amax(np.amax(abs(relDiff), 0), 0)

comp3VollEigNormalizedSteer, steerNorm = beamNew.beamformerFreq(True, 3, False, normFull, (r0, rm, kj, eigVal[:, indLow : indHigh], eigVec[:, :, indLow : indHigh]))
comp3VollEigNormalizedSteer /= steerNorm
comp3VollEigNormalizedSteerRefNumpy = beamformerRef(csm, steerFormulation3normalized)
relDiff = (comp3VollEigNormalizedSteer - comp3VollEigNormalizedSteerRefNumpy) / (comp3VollEigNormalizedSteer + comp3VollEigNormalizedSteerRefNumpy) * 2
error3VollEigNormalizedSteerAgainstNumpy = np.amax(np.amax(abs(relDiff), 0), 0)

comp3DiagEigNormalizedSteer, steerNorm = beamNew.beamformerFreq(True, 3, True, normDiag, (r0, rm, kj, eigVal[:, indLow : indHigh], eigVec[:, :, indLow : indHigh]))
comp3DiagEigNormalizedSteer /= steerNorm
comp3DiagEigNormalizedSteerRefNumpy = beamformerRef(csmRemovedDiag, steerFormulation3normalized) * normDiag
relDiff = (comp3DiagEigNormalizedSteer - comp3DiagEigNormalizedSteerRefNumpy) / (comp3DiagEigNormalizedSteer + comp3DiagEigNormalizedSteerRefNumpy) * 2
error3DiagEigNormalizedSteerAgainstNumpy = np.amax(np.amax(abs(relDiff), 0), 0)



comp4VollNormalizedSteer, steerNorm = beamNew.beamformerFreq(False, 4, False, normFull, (r0, rm, kj, csm))
comp4VollNormalizedSteer /= steerNorm
comp4VollNormalizedSteerRefNumpy = beamformerRef(csm, steerFormulation4normalized)
relDiff = (comp4VollNormalizedSteer - comp4VollNormalizedSteerRefNumpy) / (comp4VollNormalizedSteer + comp4VollNormalizedSteerRefNumpy) * 2
error4VollNormalizedSteerAgainstNumpy = np.amax(np.amax(abs(relDiff), 0), 0)

comp4DiagNormalizedSteer, steerNorm = beamNew.beamformerFreq(False, 4, True, normDiag, (r0, rm, kj, csm))
comp4DiagNormalizedSteer /= steerNorm
comp4DiagNormalizedSteerRefNumpy = beamformerRef(csmRemovedDiag, steerFormulation4normalized) * normDiag
relDiff = (comp4DiagNormalizedSteer - comp4DiagNormalizedSteerRefNumpy) / (comp4DiagNormalizedSteer + comp4DiagNormalizedSteerRefNumpy) * 2
error4DiagNormalizedSteerAgainstNumpy = np.amax(np.amax(abs(relDiff), 0), 0)

comp4VollEigNormalizedSteer, steerNorm = beamNew.beamformerFreq(True, 4, False, normFull, (r0, rm, kj, eigVal[:, indLow : indHigh], eigVec[:, :, indLow : indHigh]))
comp4VollEigNormalizedSteer /= steerNorm
comp4VollEigNormalizedSteerRefNumpy = beamformerRef(csm, steerFormulation4normalized)
relDiff = (comp4VollEigNormalizedSteer - comp4VollEigNormalizedSteerRefNumpy) / (comp4VollEigNormalizedSteer + comp4VollEigNormalizedSteerRefNumpy) * 2
error4VollEigNormalizedSteerAgainstNumpy = np.amax(np.amax(abs(relDiff), 0), 0)

comp4DiagEigNormalizedSteer, steerNorm = beamNew.beamformerFreq(True, 4, True, normDiag, (r0, rm, kj, eigVal[:, indLow : indHigh], eigVec[:, :, indLow : indHigh]))
comp4DiagEigNormalizedSteer /= steerNorm
comp4DiagEigNormalizedSteerRefNumpy = beamformerRef(csmRemovedDiag, steerFormulation4normalized) * normDiag
relDiff = (comp4DiagEigNormalizedSteer - comp4DiagEigNormalizedSteerRefNumpy) / (comp4DiagEigNormalizedSteer + comp4DiagEigNormalizedSteerRefNumpy) * 2
error4DiagEigNormalizedSteerAgainstNumpy = np.amax(np.amax(abs(relDiff), 0), 0)


### --> loojks good

#%% new beamformer - against weave
indLow = 0
indHigh = nMics

# the normalization factors as used in future with NUMBA
normFull = 1.0
normDiag = np.float64(nMics) / np.float64(nMics - 1)

# the normalization factors as used at the moment with weave
normFullWeave = np.float64(nMics ** 2)
normDiagWeave = np.float64(nMics * (nMics - 1))

#==============================================================================
# # see below why the csm can't be transposed as the input of weave 
# # (but must be transposed as input of numpy and new beamformer)
#==============================================================================
comp1Voll, steerNorm = beamNew.beamformerFreq(False, 1, False, normFull, (r0, rm, kj, csm.transpose(0,2,1)))
beamformer.r_beamfull_classic(csm, e, h, r0, rm, kj)
comp1VollRefWeave = h / normFullWeave
relDiff = (comp1Voll - comp1VollRefWeave) / (comp1Voll + comp1VollRefWeave) * 2
error1VollAgainstWeave = np.amax(np.amax(abs(relDiff), 0), 0)

comp1Diag, steerNorm = beamNew.beamformerFreq(False, 1, True, normDiag, (r0, rm, kj, csm.transpose(0,2,1)))
beamformer.r_beamdiag_classic(csm, e, h, r0, rm, kj)
comp1DiagRefWeave = h / normDiagWeave
relDiff = (comp1Diag - comp1DiagRefWeave) / (comp1Diag + comp1DiagRefWeave) * 2
error1DiagAgainstWeave = np.amax(np.amax(abs(relDiff), 0), 0)

comp1VollEig, steerNorm = beamNew.beamformerFreq(True, 1, False, normFull, (r0, rm, kj, eigValTrans[:, indLow : indHigh], eigVecTrans[:, :, indLow : indHigh]))
beamformer.r_beamfull_os_classic(e, h, r0, rm, kj, eigVal, eigVec, indLow, indHigh)
comp1VollEigRefWeave = h / normFullWeave
relDiff = (comp1VollEig - comp1VollEigRefWeave) / (comp1VollEig + comp1VollEigRefWeave) * 2
error1VollEigAgainstWeave = np.amax(np.amax(abs(relDiff), 0), 0)

comp1DiagEig, steerNorm = beamNew.beamformerFreq(True, 1, True, normDiag, (r0, rm, kj, eigValTrans[:, indLow : indHigh], eigVecTrans[:, :, indLow : indHigh]))
beamformer.r_beamdiag_os_classic(e, h, r0, rm, kj, eigVal, eigVec, indLow, indHigh)
comp1DiagEigRefWeave = h / normDiagWeave
relDiff = (comp1DiagEig - comp1DiagEigRefWeave) / (comp1DiagEig + comp1DiagEigRefWeave) * 2
error1DiagEigAgainstWeave = np.amax(np.amax(abs(relDiff), 0), 0)



comp2Voll, steerNorm = beamNew.beamformerFreq(False, 2, False, normFull, (r0, rm, kj, csm.transpose(0,2,1)))
beamformer.r_beamfull_inverse(csm, e, h, r0, rm, kj)
comp2VollRefWeave = h / normFullWeave
relDiff = (comp2Voll - comp2VollRefWeave) / (comp2Voll + comp2VollRefWeave) * 2
error2VollAgainstWeave = np.amax(np.amax(abs(relDiff), 0), 0)

comp2Diag, steerNorm = beamNew.beamformerFreq(False, 2, True, normDiag, (r0, rm, kj, csm.transpose(0,2,1)))
beamformer.r_beamdiag_inverse(csm, e, h, r0, rm, kj)
comp2DiagRefWeave = h / normDiagWeave
relDiff = (comp2Diag - comp2DiagRefWeave) / (comp2Diag + comp2DiagRefWeave) * 2
error2DiagAgainstWeave = np.amax(np.amax(abs(relDiff), 0), 0)

comp2VollEig, steerNorm = beamNew.beamformerFreq(True, 2, False, normFull, (r0, rm, kj, eigValTrans[:, indLow : indHigh], eigVecTrans[:, :, indLow : indHigh]))
beamformer.r_beamfull_os_inverse(e, h, r0, rm, kj, eigVal, eigVec, indLow, indHigh)
comp2VollEigRefWeave = h / normFullWeave
relDiff = (comp2VollEig - comp2VollEigRefWeave) / (comp2VollEig + comp2VollEigRefWeave) * 2
error2VollEigAgainstWeave = np.amax(np.amax(abs(relDiff), 0), 0)

comp2DiagEig, steerNorm = beamNew.beamformerFreq(True, 2, True, normDiag, (r0, rm, kj, eigValTrans[:, indLow : indHigh], eigVecTrans[:, :, indLow : indHigh]))
beamformer.r_beamdiag_os_inverse(e, h, r0, rm, kj, eigVal, eigVec, indLow, indHigh)
comp2DiagEigRefWeave = h / normDiagWeave
relDiff = (comp2DiagEig - comp2DiagEigRefWeave) / (comp2DiagEig + comp2DiagEigRefWeave) * 2
error2DiagEigAgainstWeave = np.amax(np.amax(abs(relDiff), 0), 0)



comp3Voll, steerNorm = beamNew.beamformerFreq(False, 3, False, normFull, (r0, rm, kj, csm.transpose(0,2,1)))
beamformer.r_beamfull(csm, e, h, r0, rm, kj)
comp3VollRefWeave = h / normFullWeave
relDiff = (comp3Voll - comp3VollRefWeave) / (comp3Voll + comp3VollRefWeave) * 2
error3VollAgainstWeave = np.amax(np.amax(abs(relDiff), 0), 0)

comp3Diag, steerNorm = beamNew.beamformerFreq(False, 3, True, normDiag, (r0, rm, kj, csm.transpose(0,2,1)))
beamformer.r_beamdiag(csm, e, h, r0, rm, kj)
comp3DiagRefWeave = h / normDiagWeave
relDiff = (comp3Diag - comp3DiagRefWeave) / (comp3Diag + comp3DiagRefWeave) * 2
error3DiagAgainstWeave = np.amax(np.amax(abs(relDiff), 0), 0)

comp3VollEig, steerNorm = beamNew.beamformerFreq(True, 3, False, normFull, (r0, rm, kj, eigValTrans[:, indLow : indHigh], eigVecTrans[:, :, indLow : indHigh]))
beamformer.r_beamfull_os(e, h, r0, rm, kj, eigVal, eigVec, indLow, indHigh)
comp3VollEigRefWeave = h / normFullWeave
relDiff = (comp3VollEig - comp3VollEigRefWeave) / (comp3VollEig + comp3VollEigRefWeave) * 2
error3VollEigAgainstWeave = np.amax(np.amax(abs(relDiff), 0), 0)

comp3DiagEig, steerNorm = beamNew.beamformerFreq(True, 3, True, normDiag, (r0, rm, kj, eigValTrans[:, indLow : indHigh], eigVecTrans[:, :, indLow : indHigh]))
beamformer.r_beamdiag_os(e, h, r0, rm, kj, eigVal, eigVec, indLow, indHigh)
comp3DiagEigRefWeave = h / normDiagWeave
relDiff = (comp3DiagEig - comp3DiagEigRefWeave) / (comp3DiagEig + comp3DiagEigRefWeave) * 2
error3DiagEigAgainstWeave = np.amax(np.amax(abs(relDiff), 0), 0)



comp4Voll, steerNorm = beamNew.beamformerFreq(False, 4, False, normFull, (r0, rm, kj, csm.transpose(0,2,1)))
beamformer.r_beamfull_3d(csm, e, h, r0, rm, kj)
comp4VollRefWeave = h / normFullWeave
relDiff = (comp4Voll - comp4VollRefWeave) / (comp4Voll + comp4VollRefWeave) * 2
error4VollAgainstWeave = np.amax(np.amax(abs(relDiff), 0), 0)

comp4Diag, steerNorm = beamNew.beamformerFreq(False, 4, True, normDiag, (r0, rm, kj, csm.transpose(0,2,1)))
beamformer.r_beamdiag_3d(csm, e, h, r0, rm, kj)
comp4DiagRefWeave = h / normDiagWeave
relDiff = (comp4Diag - comp4DiagRefWeave) / (comp4Diag + comp4DiagRefWeave) * 2
error4DiagAgainstWeave = np.amax(np.amax(abs(relDiff), 0), 0)

comp4VollEig, steerNorm = beamNew.beamformerFreq(True, 4, False, normFull, (r0, rm, kj, eigValTrans[:, indLow : indHigh], eigVecTrans[:, :, indLow : indHigh]))
beamformer.r_beamfull_os_3d(e, h, r0, rm, kj, eigVal, eigVec, indLow, indHigh)
comp4VollEigRefWeave = h / normFullWeave
relDiff = (comp4VollEig - comp4VollEigRefWeave) / (comp4VollEig + comp4VollEigRefWeave) * 2
error4VollEigAgainstWeave = np.amax(np.amax(abs(relDiff), 0), 0)

comp4DiagEig, steerNorm = beamNew.beamformerFreq(True, 4, True, normDiag, (r0, rm, kj, eigValTrans[:, indLow : indHigh], eigVecTrans[:, :, indLow : indHigh]))
beamformer.r_beamdiag_os_3d(e, h, r0, rm, kj, eigVal, eigVec, indLow, indHigh)
comp4DiagEigRefWeave = h / normDiagWeave
relDiff = (comp4DiagEig - comp4DiagEigRefWeave) / (comp4DiagEig + comp4DiagEigRefWeave) * 2
error4DiagEigAgainstWeave = np.amax(np.amax(abs(relDiff), 0), 0)

### --> looks good

#%% Use only float32 for all calculations in beamformer
erg64 = beamformerRef(csm, steerFormulation1)
erg32 = beamformerRef(np.complex64(csm), np.complex64(steerFormulation1))
relDiff = np.amax(np.amax((erg64 - erg32) / (erg64 + erg32), 0), 0)

### --> there will be massive errors, if one uses only float32 datatypes!

#%% weave reacts strange
#==============================================================================
# There is a difference when giving weave the transpose of the csm (and 'correct'-algo gets the normal csm), 
# or giving the 'correct' algorithm the csm-transpose and the weave gets the untransposed csm.
# Theoretically this should produce the same results. Maybe the problem is the ffast03 compiling option in weave?
#==============================================================================
normalizationFull = np.float64(nMics ** 2)

numpyResult = vectorized(csm.transpose(0,2,1), e, h, r0, rm, kj, normalizationFull)  
neu, steerNorm = beamNew.beamformerFreq(False, 1, False, normFull, (r0, rm, kj, csm.transpose(0,2,1)))
beamformer.r_beamfull_classic(csm, e, h, r0, rm, kj)
weave = h / normalizationFull

relDiff = (neu - weave) / (neu + weave) * 2
errorNeuVsWeave = np.amax(np.amax(abs(relDiff), 0), 0)

relDiff = (neu - numpyResult) / (neu + numpyResult) * 2
errorNeuVsNumpy = np.amax(np.amax(abs(relDiff), 0), 0)

relDiff = (weave - numpyResult) / (weave + numpyResult) * 2
errorNPvsWeave = np.amax(np.amax(abs(relDiff), 0), 0)



import unittest

from os.path import join

import numpy as np

from acoular import config
config.global_caching = 'none'

from acoular import WNoiseGenerator, MovingPointSource, WriteH5, Trajectory, MicGeom,\
    RectGrid, MaskedTimeSamples, SteeringVector, TimeAverage, BeamformerTimeSqTraj, BeamformerTimeTraj,\
    BeamformerCleantTraj, BeamformerCleantSqTraj, BeamformerTime, BeamformerTimeSq,\
    BeamformerCleant, BeamformerCleantSq

# if this flag is set to True, new time data will be simulated and
WRITE_NEW_REFERENCE_DATA = False
# new beamformer results are generated for comparison during testing. Should always be False. Only set to
# true, if it is necessary to recalculate the data, due to wanted changes of the Beamformers (or MovingPointSource).

# Parameters
FNAME = join('reference_data','beamformer_traj_time_data.h5')
SFREQ = 6000
SPEED = 10  # km/h
SEED = 1
D = .5
SOURCE_POS = (0.0, 0.0, D)
passby_dist = .5  # distance that the source is passing in front of array
CONV_AMP = True

# create linear mic geom
MGEOM = MicGeom()
N = 5
L = .5
MGEOM.mpos_tot = np.zeros((3, N), dtype=np.float64)
win = np.sin(np.arange(N)*np.pi/(N-1))
b = 0.4
MGEOM.mpos_tot[0] = np.linspace(-L, L, N)*(1-b)/(win*b+1-b)

# Monopole Trajectory
t_passby = passby_dist/SPEED/3.6
nsamples = int(t_passby*SFREQ)
TRAJ = Trajectory()  # source center
TRAJ.points[0] = (-passby_dist/2 + SOURCE_POS[0], SOURCE_POS[1], SOURCE_POS[2])
TRAJ.points[t_passby] = (+passby_dist/2, SOURCE_POS[1], SOURCE_POS[2])


def create_test_time_data(nsamples):
    """
    creates test data for a single moving monopole emitting white noise
    
    Parameters
    ----------
    speed : float
        source velocity in m/s.

    Returns
    -------
    None.

    """
    n1 = WNoiseGenerator(sample_freq=SFREQ, numsamples=nsamples, seed=SEED)
    p1 = MovingPointSource(signal=n1, mics=MGEOM,
                           trajectory=TRAJ, conv_amp=CONV_AMP)
    wh5 = WriteH5(source=p1, name=FNAME)
    print(50*"#")
    print(f"create {FNAME} ...")
    print(f"num samples: {nsamples}, pass-by time: {t_passby}s")
    print(50*"#")
    wh5.save()

def get_beamformer_traj_result(Beamformer, num=32):
    """
    returns the result for a given Beamformer class
    
    Parameters
    ----------
    Beamformer : cls
        trajectory beamformer.
    num : int, optional
        number of samples to return. The default is 32.

    Returns
    -------
    array
        first block returned by the trajectory beamformers result() function.

    """
    ## with moving grid
    ts = MaskedTimeSamples(name=FNAME)
    gMoving = RectGrid(x_min=-.1, x_max=.1, y_min=0, y_max=0, z=0,
                       increment=.1)
    stMoving = SteeringVector(grid=gMoving, mics=MGEOM)
    bt = Beamformer(source=ts, trajectory=TRAJ, steer=stMoving)
    if hasattr(bt,'n_iter'):
        bt.n_iter = 2
    return next(bt.result(num)).astype(np.float32)

def get_beamformer_time_result(Beamformer, num=32):
    """
    returns the result for a given time Beamformer class
    
    Parameters
    ----------
    Beamformer : cls
        time beamformer.
    num : int, optional
        number of samples to return. The default is 32.

    Returns
    -------
    array
        first block returned by the time beamformers result() function.

    """
    ## with moving grid
    ts = MaskedTimeSamples(name=FNAME)
    gfixed = RectGrid(x_min=-.1, x_max=.1, y_min=0, y_max=0, z=D,
                       increment=.1)
    stfixed = SteeringVector(grid=gfixed, mics=MGEOM)
    bt = Beamformer(source=ts,steer=stfixed)
    if hasattr(bt,'n_iter'):
        bt.n_iter = 2
    return next(bt.result(num)).astype(np.float32)

class BeamformerTimeTest(unittest.TestCase):
    """
    A simple test case that varifies that the results of trajectory beamformers 
    and fixed focus beamformers in time domain are not changing across different
    versions of code.
    """

    traj_beamformers = [BeamformerTimeTraj, BeamformerTimeSqTraj,
                        BeamformerCleantTraj, BeamformerCleantSqTraj]

    time_beamformers = [BeamformerTime, BeamformerTimeSq,
                        BeamformerCleant, BeamformerCleantSq]
    
    def test_beamformer_traj_result(self):
        """compare results of trajectory beamformers against previous
        results from .h5 file"""
        for beamformer in self.traj_beamformers:
            with self.subTest(beamformer.__name__):
                name = join('reference_data',f'{beamformer.__name__}.npy')
                actual_data = get_beamformer_traj_result(beamformer)
                if WRITE_NEW_REFERENCE_DATA:
                    np.save(name,actual_data)
                ref_data = np.load(name)
                np.testing.assert_allclose(actual_data, ref_data, rtol=5e-5, atol=5e-6)

    def test_beamformer_time_result(self):
        """compare results of time beamformers with fixed focus against previous
        results from .h5 file"""
        for beamformer in self.time_beamformers:
            with self.subTest(beamformer.__name__):
                name = join('reference_data',f'{beamformer.__name__}.npy')
                actual_data = get_beamformer_time_result(beamformer)
                if WRITE_NEW_REFERENCE_DATA:
                    np.save(name,actual_data)
                ref_data = np.load(name)
                np.testing.assert_allclose(actual_data, ref_data, rtol=5e-5, atol=5e-6)

    
if __name__ == '__main__':
    unittest.main()




from os import path
from numpy import array, round
import matplotlib.pyplot as plt
from acoular import __file__ as bpath, MicGeom, WNoiseGenerator, PointSource,\
    Mixer, WriteH5, TimeSamples, PowerSpectra, RectGrid, SteeringVector,\
    BeamformerCleansc, L_p, ImportGrid,CircSector
from acoular.tools import MetricEvaluator

# set up the parameters
sfreq = 51200 
duration = 1
nsamples = duration*sfreq
micgeofile = path.join(path.split(bpath)[0],'xml','array_64.xml')
h5savefile = 'three_sources.h5'

# generate test data, in real life this would come from an array measurement
mg = MicGeom( from_file=micgeofile )
n1 = WNoiseGenerator( sample_freq=sfreq, numsamples=nsamples, seed=1 )
n2 = WNoiseGenerator( sample_freq=sfreq, numsamples=nsamples, seed=2, rms=0.7 )
n3 = WNoiseGenerator( sample_freq=sfreq, numsamples=nsamples, seed=3, rms=0.5 )
p1 = PointSource( signal=n1, mics=mg,  loc=(-0.1,-0.1,0.3) )
p2 = PointSource( signal=n2, mics=mg,  loc=(0.15,0,0.3) )
p3 = PointSource( signal=n3, mics=mg,  loc=(0,0.1,0.3) )
pa = Mixer( source=p1, sources=[p2,p3] )
wh5 = WriteH5( source=pa, name=h5savefile )
wh5.save()

# analyze the data and generate map

ts = TimeSamples( name=h5savefile )
ps = PowerSpectra( time_data=ts, block_size=128, window='Hanning' )

rg = RectGrid( x_min=-0.2, x_max=0.2, y_min=-0.2, y_max=0.2, z=0.3, \
increment=0.01 )
st = SteeringVector(grid=rg, mics=mg, ref = 1.)

bb = BeamformerCleansc( freq_data=ps, steer=st )
pm = bb.synthetic( 8000, 0 )
Lm = L_p( pm )

# evaluate the results 
target_grid = ImportGrid(
    gpos_file= array([
    list(p1.loc),
    list(p2.loc),
    list(p3.loc)
]).T)

nfft = ps.fftfreq().shape[0]

target_data = array([
    [n1.rms**2/nfft],
    [n2.rms**2/nfft],
    [n3.rms**2/nfft]]).T

mv = MetricEvaluator(
    sector = CircSector(r=0.05*mg.aperture),
    grid = rg,
    data = pm.reshape((1,-1)),
    target_grid = target_grid,
    target_data = target_data,
)

# plot the data

plt.figure()
# show map
plt.imshow( Lm.T, origin='lower', vmin=Lm.max()-10, extent=rg.extend(), \
interpolation='none')
# plot sectors
ax = plt.gca()
for j, sector in enumerate(mv.sectors):
    ax.add_patch(plt.Circle((sector.x, sector.y), 
                sector.r, color="black",fill=False))
    # annotate specific level error below circles
    plt.annotate(r'$\Delta L_{p,e,s}$='+ str(round(mv.get_specific_level_error()[0,j],2)) + ' dB',
                xy=(sector.x-0.1,sector.y-sector.r-0.01), color='white')
# annotate overall level error
plt.annotate(r'$\Delta L_{p,e,o}$='+ str(round(mv.get_overall_level_error()[0],2)) + ' dB',
            xy=(0.05,0.95),xycoords='axes fraction', color='white')
plt.annotate(r'$\Delta L_{p,e,i}$='+ str(round(mv.get_inverse_level_error()[0],2)) + ' dB',
            xy=(0.6,0.95),xycoords='axes fraction', color='white')    
plt.colorbar()



import unittest
import acoular as ac
ac.config.global_caching = "none"
import numpy as np
from test_grid import GridTest
from functools import partial

class TestIntegrate(unittest.TestCase):
    
    f = [1000,2000]

    @staticmethod
    def get_sector_args():
        # for later testing condition: sector only includes (0,0,1) point 
        return {'RectGrid': np.array([0,0,0.2]),
                'RectGrid3D' : np.array([0,0,1,0,0,1]),}

    def get_beamformer(self, grid):
        rng1 = np.random.RandomState(1)
        src_pos = np.array([[0],[0],[1]])
        mics = ac.MicGeom(mpos_tot=rng1.normal(size=3*5).reshape((3,5)))
        steer = ac.SteeringVector(
                        grid=ac.ImportGrid(
                            gpos_file=src_pos), 
                        mics=mics)
        H = np.empty((len(self.f),mics.num_mics, 1),dtype=complex)
        for i,_f in enumerate(self.f): # calculate only the indices that are needed
            H[i] = steer.transfer(_f).T # transfer functions
        csm = H@H.swapaxes(2,1).conjugate()
        freq_data = ac.PowerSpectraImport(csm = csm, frequencies=self.f)
        steer.grid = grid
        return ac.BeamformerBase(freq_data=freq_data, steer=steer)

    def test_sector_class_integration_functional(self):
        for sector in GridTest.get_sector_classes():
            for grid in GridTest.get_all_grids():
                bf = self.get_beamformer(grid)
                with self.subTest(
                    f"Grid: {grid.__class__.__name__} Sector: {sector.__class__.__name__}"):
                    for i,f in enumerate(self.f):
                        bf_res = bf.synthetic(f)
                        bf_max = bf_res.max()
                        integration_res = ac.integrate(
                            data=bf_res,sector=sector,grid=grid)
                        self.assertEqual(integration_res.shape, ())
                        self.assertEqual(integration_res, bf_max)

    def test_sector_class_integration_class(self):
        for sector in GridTest.get_sector_classes():
            for grid in GridTest.get_all_grids():
                bf = self.get_beamformer(grid)
                with self.subTest(
                    f"Grid: {grid.__class__.__name__} Sector:{sector.__class__.__name__}"):                   
                    for i,f in enumerate(self.f):
                        bf_res = bf.synthetic(f)
                        bf_max = bf_res.max()
                        integration_res = bf.integrate(sector)
                        self.assertEqual(integration_res.shape, (len(self.f),))
                        self.assertEqual(integration_res[i], bf_max)

    def test_sector_args_integration_functional(self):
        for grid in GridTest.get_all_grids():
            bf = self.get_beamformer(grid)
            with self.subTest(
                f"Grid: {grid.__class__.__name__}"):                   
                for i,f in enumerate(self.f):
                    sector = self.get_sector_args().get(grid.__class__.__name__)
                    bf_res = bf.synthetic(f)
                    bf_max = bf_res.max()
                    if sector is None: # not allowed grid for simple sector args
                        sector = np.array([0,0,0.2]) # some random circ sector arguments
                        integrate = partial(
                            ac.integrate, data=bf_res, grid=grid, sector=sector)
                        self.assertRaises(NotImplementedError, integrate)
                    else:
                        integration_res = ac.integrate(
                            data=bf_res,sector=sector,grid=grid)
                        self.assertEqual(integration_res.shape, ())
                        self.assertEqual(integration_res, bf_max)

    def test_sector_args_integration_class(self):
        for grid in GridTest.get_all_grids():
            bf = self.get_beamformer(grid)
            with self.subTest(
                f"Grid: {grid.__class__.__name__}"):                   
                for i,f in enumerate(self.f):
                    sector = self.get_sector_args().get(grid.__class__.__name__)
                    bf_res = bf.synthetic(f)
                    bf_max = bf_res.max()
                    if sector is None: # not allowed grid for simple sector args
                        sector = np.array([0,0,0.2]) # some random circ sector arguments
                        integrate = partial(
                            bf.integrate, sector=sector)
                        self.assertRaises(NotImplementedError, integrate)
                    else:
                        integration_res = bf.integrate(sector)
                        self.assertEqual(integration_res.shape, (len(self.f),))
                        self.assertEqual(integration_res[i], bf_max)  

if __name__ == "__main__":

    unittest.main()



#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Thu Jul  6 16:16:03 2017

@author: tomgensch
"""
#import os
#print(os.path.dirname(os.path.realpath(__file__)))

import numpy as np
from cythonBeamformer import beamformerCython, beamformerCythonNOTparallel

nMics = 32
nGridPoints = 100
nFreqs = 10

csm = np.random.rand(nFreqs, nMics, nMics) + 1j*np.random.rand(nFreqs, nMics, nMics)
for cntFreqs in range(nFreqs):
    csm[cntFreqs, :, :] += csm[cntFreqs, :, :].T.conj()  # zu Hermitischer Matrix machen
r0 = np.random.rand(nGridPoints) #abstand aufpunkte-arraymittelpunkt
rm = np.random.rand(nGridPoints, nMics) #abstand aufpunkte-arraymikrofone
kj = np.zeros(nFreqs) + 1j*np.random.rand(nFreqs) # 


#csm = np.ones((nFreqs, nMics, nMics), 'D')
#r0 = np.ones(nGridPoints) * .2
#rm = np.array([[.2, .3]])#rm = np.ones((nGridPoints, nMics)) * .24
#kj = np.zeros(nFreqs) + 1j


print('Sequentiel')
outputWithoutMP = beamformerCythonNOTparallel(csm, r0, rm, kj)

print('Parallel')
outputMP = beamformerCython(csm, r0, rm, kj)

outputMit = np.array(outputMP)
outputOhne = np.array(outputWithoutMP)




#==============================================================================
# Bei Intel gibt es mit dem setuo-file aus der Doku keine probleme.
# Die differenz zwischen den methoden und der referenz ist dann 0 (also wirklich 0, nicht e-12)
# Allerdings ist parallel viel lamngsamer!
#==============================================================================





#e1Mit = np.array(e1MP)
#e1Ohne = np.array(e1OhneMP)


## Referenz
beamformOutput = np.zeros((nFreqs, nGridPoints), np.float64)
e1 = np.zeros((nMics), np.complex128)

for cntFreqs in xrange(nFreqs):  # laeuft z.Z. nur einmal durch
    kjj = kj[cntFreqs].imag   ### könnte man vielleicht reinnehmen um unten nicht immer den realteilnehmen zu müssen
    for cntGrid in xrange(nGridPoints):
        rs = 0
        r01 = r0[cntGrid]
        
        # Erzeugen der Steering-Vectoren
        for cntMics in xrange(nMics):
            rm1 = rm[cntGrid, cntMics]
            rs += 1.0 / (rm1**2)
            temp3 = (kjj * (rm1 - r01))  # der Float bewirkt die Abweichung vom Originalergebnis um ca 10^-8
            e1[cntMics] = (np.cos(temp3) - 1j * np.sin(temp3)) * rm1
        rs = r01 ** 2
        
        # Berechnen der Matrix multiplikation
        temp1 = 0.0
        for cntMics in xrange(nMics):
            temp2 = 0.0
            for cntMics2 in xrange(cntMics):
                temp2 += csm[cntFreqs, cntMics2, cntMics] * e1[cntMics2]  # nicht ganz sicher ob richtig
            temp1 += 2 * (temp2 * e1[cntMics].conjugate()).real
            temp1 += (csm[cntFreqs, cntMics, cntMics] * np.conjugate(e1[cntMics]) * e1[cntMics]).real
        
        beamformOutput[cntFreqs, cntGrid] = (temp1 / rs).real
    

#diffE1MitOpenMP = max(abs(e1 - e1Mit))
#diffE1OhneOpenMP = max(abs(e1 - e1Ohne))
##diffE1BothCythonMethods = max([abs(e1Mit[cnt] - e1Ohne[cnt]) for cnt in range(len(e1Ohne))])
#diffE1BothCythonMethods = np.amax(abs(e1Mit - e1Ohne), axis=0)
#diffE1BothCythonMethods = np.amax(diffE1BothCythonMethods, axis=0)

diffOutputBothMethods = np.amax(abs(outputMit - outputOhne), axis=0)
diffOutputBothMethods = np.amax(abs(diffOutputBothMethods), axis=0)

diffOutputOpenMP = np.amax(abs(outputMit - beamformOutput), axis=0)
diffOutputOpenMP = np.amax(abs(diffOutputOpenMP), axis=0)

diffOutputWithoutOpenMP = np.amax(abs(outputOhne - beamformOutput), axis=0)
diffOutputWithoutOpenMP = np.amax(abs(diffOutputWithoutOpenMP), axis=0)

print(diffOutputBothMethods)


# -*- coding: utf-8 -*-
#pylint: disable-msg=E0611,C0111,C0103,R0901,R0902,R0903,R0904,W0232
#------------------------------------------------------------------------------
# Copyright (c) Acoular Development Team.
#------------------------------------------------------------------------------

# imports from other packages
from __future__ import print_function
from traits.api import HasPrivateTraits, Bool, Str, Dict, Instance, Delegate
from os import path, mkdir, environ, listdir
from weakref import WeakValueDictionary
import gc

from .configuration import Config, config
from .h5files import _get_cachefile_class

class H5cache_class(HasPrivateTraits):
    """
    Cache class that handles opening and closing 'tables.File' objects
    """

    config = Instance(Config)

    cache_dir = Delegate('config')    

    busy = Bool(False)
    
    open_files = WeakValueDictionary()
    
    openFileReferenceCount = dict()
    
    def _idle_if_busy(self):
        while self.busy:
            pass

    def open_cachefile(self,cacheFileName,mode):
        File = _get_cachefile_class()
        return File(path.join(self.cache_dir, cacheFileName), mode)
    
    def close_cachefile(self,cachefile):
        self.openFileReferenceCount.pop(get_basename(cachefile))
        cachefile.close()
        
    def get_filename(self,file):
        File = _get_cachefile_class()
        if isinstance(file, File): 
            return get_basename(file)
        else:
            return 0

    def get_open_cachefiles(self):
        try:
            return self.open_files.itervalues()
        except AttributeError:
            return iter(self.open_files.values())

    def close_unreferenced_cachefiles(self):
        for openCacheFile in self.get_open_cachefiles():
            if not self.is_reference_existent(openCacheFile):
#                print("close unreferenced File:",get_basename(openCacheFile))
                self.close_cachefile(openCacheFile)

    def is_reference_existent(self,file):
        existFlag = False
        # inspect all refererres to the file object
        gc.collect() #clear garbage before collecting referrers
        for ref in gc.get_referrers(file):
            # does the file object have a referrer that has a 'h5f' 
            # attribute?
            if isinstance(ref,dict) and 'h5f' in ref:
                # file is still referred, must not be closed
                existFlag = True
                break
        return existFlag

    def is_cachefile_existent(self,cacheFileName):
        if cacheFileName in listdir(self.cache_dir):
            return True
        else:
            return False

    def _increase_file_reference_counter(self, cacheFileName):
        self.openFileReferenceCount[cacheFileName] = self.openFileReferenceCount.get(cacheFileName, 0) + 1

    def _decrease_file_reference_counter(self, cacheFileName):
        self.openFileReferenceCount[cacheFileName] = self.openFileReferenceCount[cacheFileName] - 1

    def _print_open_files(self):
        print(list(self.openFileReferenceCount.items()))

    def get_cache_file( self, obj, basename, mode='a' ):
        '''
        returns pytables .h5 file to h5f trait of calling object for caching
        '''        
        
        self._idle_if_busy() # 
        self.busy = True

        cacheFileName = basename + '_cache.h5'
        objFileName = self.get_filename(obj.h5f)
        
        if objFileName:
            if objFileName == cacheFileName: 
                self.busy = False
                return
            else: # in case the base name has changed ( different source ) 
                self._decrease_file_reference_counter(objFileName)

        if cacheFileName not in self.open_files: # or tables.file._open_files.filenames
            if (
                config.global_caching == 'readonly' 
                and not self.is_cachefile_existent(cacheFileName)
                ): # condition ensures that cachefile is not created in readonly mode
                obj.h5f = None
                self.busy = False
#                self._print_open_files()   
                return
            else:
                if config.global_caching == 'readonly': mode = 'r'
                f = self.open_cachefile(cacheFileName,mode)
                self.open_files[cacheFileName] = f
        
        obj.h5f = self.open_files[cacheFileName]
        self._increase_file_reference_counter(cacheFileName)
        
        # garbage collection
        self.close_unreferenced_cachefiles()
        
        self.busy = False
        self._print_open_files()   

H5cache = H5cache_class(config=config)

def get_basename(file): 
    return path.basename(file.filename)




# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
# Copyright (c) Acoular Development Team.
# ------------------------------------------------------------------------------

from numpy import empty, float32, int32
import numpy.random as random
import acoular
from numba import set_num_threads
#set_num_threads(1)

# set up test case
nMics = 64
nSamples = 2048
gridSize = 2048
num = 512
data = random.normal(size=(nSamples, nMics))
offsets = random.randint(low=0, high=nSamples / 2, size=(gridSize, nMics))
ifactor2 = random.uniform(low=0, high=1.0, size=(gridSize, nMics))
steeramp = random.uniform(low=0, high=1.0, size=(gridSize, nMics))
offsets5 = random.randint(low=0, high=nSamples / 2, size=(num, gridSize, nMics))
ifactor25 = random.uniform(low=0, high=1.0, size=(num, gridSize, nMics))
steeramp5 = random.uniform(low=0, high=1.0, size=(num, gridSize, nMics))
out = empty((num, gridSize))
autopower = empty((num, gridSize))


def test_delayandsum4():
    acoular.tfastfuncs._delayandsum4(data, offsets, ifactor2, steeramp, out, autopower)


def test_delayandsum5():
    acoular.tfastfuncs._delayandsum5(
        data, offsets5, ifactor25, steeramp5, out, autopower
    )


data32 = data.astype(float32)
offsets532 = offsets5.astype(int32)
ifactor2532 = ifactor25.astype(float32)
steeramp532 = steeramp5.astype(float32)
out32 = out.astype(float32)
autopower32 = autopower.astype(float32)


def test_delayandsum532():
    acoular.tfastfuncs._delayandsum5(
        data32, offsets532, ifactor2532, steeramp532, out32, autopower32
    )

rm = ifactor25.copy()
delays = ifactor25.copy()
interp = ifactor25.copy()
index = offsets5.copy()

def test_delays():
    acoular.tfastfuncs._delays(
        rm, 343., interp, index
    )

rm32 = rm.astype(float32)
interp32 = interp.astype(float32)
index32 = index.astype(int32)

def test_delays32():
    acoular.tfastfuncs._delays(
        rm32, float32(343.), interp32, index32
    )

r0 = random.uniform(low=0, high=1.0, size=(num, gridSize))

def test_steer_I():
    acoular.tfastfuncs._steer_I(
        rm, r0, steeramp5
    )

def test_steer_II():
    acoular.tfastfuncs._steer_II(
        rm, r0, steeramp5
    )

def test_steer_III():
    acoular.tfastfuncs._steer_III(
        rm, r0, steeramp5
    )

def test_steer_IV():
    acoular.tfastfuncs._steer_IV(
        rm, r0, steeramp5
    )


# -*- coding: utf-8 -*-
#pylint: disable-msg=E0611, E1101, C0103, R0901, R0902, R0903, R0904, W0232
#------------------------------------------------------------------------------
# Copyright (c) Acoular Development Team.
#------------------------------------------------------------------------------
"""Implements beamformers in the time domain.

.. autosummary::
    :toctree: generated/

    BeamformerTime
    BeamformerTimeTraj
    BeamformerTimeSq
    BeamformerTimeSqTraj
    BeamformerCleant
    BeamformerCleantTraj
    BeamformerCleantSq
    BeamformerCleantSqTraj
    IntegratorSectorTime
"""

# imports from other packages
from __future__ import print_function, division
from numpy import array, newaxis, empty, sqrt, arange, r_, zeros, \
histogram, unique, dot, where, s_ , sum, isscalar, full, ceil, argmax,\
interp,concatenate, float32, float64, int32, int64
from numpy.linalg import norm
from traits.api import Float, CArray, Property, Trait, Bool, Delegate, \
cached_property, List, Instance, Range, Int, Enum
from traits.trait_errors import TraitError
from warnings import warn

# acoular imports
from .internal import digest
from .grids import RectGrid
from .trajectory import Trajectory
from .tprocess import TimeInOut
from .fbeamform import SteeringVector, L_p
from .tfastfuncs import _delayandsum4, _delayandsum5, \
    _steer_I, _steer_II, _steer_III, _steer_IV, _delays, _modf


def const_power_weight( bf ):
    """
    Internal helper function for :class:`BeamformerTime`
    
    Provides microphone weighting 
    to make the power per unit area of the
    microphone array geometry constant.
    
    Parameters
    ----------
    bf: :class:`BeamformerTime` object
        
          
    Returns
    -------
    array of floats
        The weight factors.
    """

    r = bf.steer.env._r(zeros((3, 1)), bf.steer.mics.mpos) # distances to center
    # round the relative distances to one decimal place
    r = (r/r.max()).round(decimals=1)
    ru, ind = unique(r, return_inverse=True)
    ru = (ru[1:]+ru[:-1])/2
    count, bins = histogram(r, r_[0, ru, 1.5*r.max()-0.5*ru[-1]])
    bins *= bins
    weights = sqrt((bins[1:]-bins[:-1])/count)
    weights /= weights.mean()
    return weights[ind]

# possible choices for spatial weights
possible_weights = {'none':None, 
                    'power':const_power_weight}


class BeamformerTime( TimeInOut ):
    """
    Provides a basic time domain beamformer with time signal output
    for a spatially fixed grid.
    """


    # Instance of :class:`~acoular.fbeamform.SteeringVector` or its derived classes
    # that contains information about the steering vector. This is a private trait.
    # Do not set this directly, use `steer` trait instead.
    _steer_obj = Instance(SteeringVector(), SteeringVector)   
    
    #: :class:`~acoular.fbeamform.SteeringVector` or derived object. 
    #: Defaults to :class:`~acoular.fbeamform.SteeringVector` object.
    steer = Property(desc="steering vector object")  
    
    def _get_steer(self):
        return self._steer_obj
    
    def _set_steer(self, steer):
        if type(steer) == SteeringVector:
            # This condition may be replaced at a later time by: isinstance(steer, SteeringVector): -- (derived classes allowed)
            self._steer_obj = steer
        elif steer in ('true level', 'true location', 'classic', 'inverse'):
            # Type of steering vectors, see also :ref:`Sarradj, 2012<Sarradj2012>`.
            warn("Deprecated use of 'steer' trait. "
                 "Please use object of class 'SteeringVector' in the future.", 
                 Warning, stacklevel = 2)
            self._steer_obj.steer_type = steer
        else:
            raise(TraitError(args=self,
                             name='steer', 
                             info='SteeringVector',
                             value=steer))

    # --- List of backwards compatibility traits and their setters/getters -----------
    
    # :class:`~acoular.environments.Environment` or derived object. 
    # Deprecated! Only kept for backwards compatibility. 
    # Now governed by :attr:`steer` trait.
    env = Property()
    
    def _get_env(self):
        return self._steer_obj.env    
    
    def _set_env(self, env):
        warn("Deprecated use of 'env' trait. ", Warning, stacklevel = 2)
        self._steer_obj.env = env
    
    # The speed of sound.
    # Deprecated! Only kept for backwards compatibility. 
    # Now governed by :attr:`steer` trait.
    c = Property()
    
    def _get_c(self):
        return self._steer_obj.env.c
    
    def _set_c(self, c):
        warn("Deprecated use of 'c' trait. ", Warning, stacklevel = 2)
        self._steer_obj.env.c = c
   
    # :class:`~acoular.grids.Grid`-derived object that provides the grid locations.
    # Deprecated! Only kept for backwards compatibility. 
    # Now governed by :attr:`steer` trait.
    grid = Property()

    def _get_grid(self):
        return self._steer_obj.grid
    
    def _set_grid(self, grid):
        warn("Deprecated use of 'grid' trait. ", Warning, stacklevel = 2)
        self._steer_obj.grid = grid
    
    # :class:`~acoular.microphones.MicGeom` object that provides the microphone locations.
    # Deprecated! Only kept for backwards compatibility. 
    # Now governed by :attr:`steer` trait
    mpos = Property()
    
    def _get_mpos(self):
        return self._steer_obj.mics
    
    def _set_mpos(self, mpos):
        warn("Deprecated use of 'mpos' trait. ", Warning, stacklevel = 2)
        self._steer_obj.mics = mpos
    
    
    # Sound travel distances from microphone array center to grid points (r0)
    # and all array mics to grid points (rm). Readonly.
    # Deprecated! Only kept for backwards compatibility. 
    # Now governed by :attr:`steer` trait
    r0 = Property()
    def _get_r0(self):
        return self._steer_obj.r0
    
    rm = Property()
    def _get_rm(self):
        return self._steer_obj.rm
    
    # --- End of backwards compatibility traits --------------------------------------

    #: Number of channels in output (=number of grid points).
    numchannels = Delegate('grid', 'size')

    #: Spatial weighting function.
    weights = Trait('none', possible_weights, 
        desc="spatial weighting function")
    # (from timedomain.possible_weights)

    # buffer with microphone time signals used for processing. Internal use 
    buffer = CArray(desc="buffer containing microphone signals")
    
    # index indicating position of current processing sample. Internal use.
    bufferIndex = Int(desc="index indicating position in buffer")

    # internal identifier
    digest = Property( 
        depends_on = ['_steer_obj.digest', 'source.digest', 'weights', '__class__'], 
        )

    @cached_property
    def _get_digest( self ):
        return digest(self)
    
    def _get_weights(self):
        if self.weights_:
            w = self.weights_(self)[newaxis]
        else:
            w = 1.0
        return w

    def _fill_buffer(self,num):
        """ generator that fills the signal buffer """
        weights = self._get_weights()
        for block in self.source.result(num):
            block *= weights
            ns = block.shape[0]
            bufferSize = self.buffer.shape[0]
            self.buffer[0:(bufferSize-ns)] = self.buffer[-(bufferSize-ns):] 
            self.buffer[-ns:,:] = block
            self.bufferIndex -= ns
            yield
         
    def result( self, num=2048 ):
        """
        Python generator that yields the *squared* deconvolved beamformer 
        output with optional removal of autocorrelation block-wise.
        
        Parameters
        ----------
        num : integer, defaults to 2048
            This parameter defines the size of the blocks to be yielded
            (i.e. the number of samples per block).
        
        Returns
        -------
        Samples in blocks of shape  \
        (num, :attr:`~BeamformerTime.numchannels`). 
            :attr:`~BeamformerTime.numchannels` is usually very \
            large (number of grid points).
            The last block may be shorter than num. \
            The output starts for signals that were emitted 
            from the grid at `t=0`.
        """
        # initialize values
        fdtype = float64
        idtype = int64
        numMics = self.steer.mics.num_mics
        n_index = arange(0,num+1)[:,newaxis]
        c = self.steer.env.c/self.source.sample_freq
        amp = empty((1,self.grid.size,numMics),dtype=fdtype)
        #delays = empty((1,self.grid.size,numMics),dtype=fdtype)
        d_index = empty((1,self.grid.size,numMics),dtype=idtype)
        d_interp2 = empty((1,self.grid.size,numMics),dtype=fdtype)
        _steer_III(self.rm[newaxis,:,:],self.r0[newaxis,:],amp)
        _delays(self.rm[newaxis,:,:], c, d_interp2, d_index)
        amp.shape = amp.shape[1:]
        #delays.shape = delays.shape[1:]
        d_index.shape = d_index.shape[1:]
        d_interp2.shape = d_interp2.shape[1:]
        maxdelay = int((self.rm/c).max())+2 + num # +2 because of interpolation
        initialNumberOfBlocks = int(ceil(maxdelay/num))
        bufferSize=initialNumberOfBlocks*num
        self.buffer = zeros((bufferSize,numMics), dtype=float)
        self.bufferIndex = bufferSize # indexing current time sample in buffer 
        fill_buffer_generator = self._fill_buffer(num)
        for _ in range(initialNumberOfBlocks):
            next(fill_buffer_generator)

        # start processing
        flag = True
        while flag:
            samplesleft = self.buffer.shape[0]-self.bufferIndex
            if samplesleft-maxdelay <= 0:
                num += samplesleft-maxdelay
                maxdelay += samplesleft-maxdelay
                n_index = arange(0,num+1)[:,newaxis]
                flag=False
            # init step
            p_res = array(
                self.buffer[self.bufferIndex:self.bufferIndex+maxdelay,:])
            Phi, autopow = self.delay_and_sum(num,p_res,d_interp2,d_index,amp)
            if 'Cleant' not in self.__class__.__name__:
                if 'Sq' not in self.__class__.__name__:
                    yield Phi[:num]
                elif self.r_diag:                 
                    yield (Phi[:num]**2 - autopow[:num]).clip(min=0)
                else:
                    yield Phi[:num]**2
            else:
                Gamma = zeros(Phi.shape)
                Gamma_autopow = zeros(Phi.shape)
                J = 0
                # deconvolution 
                while (J < self.n_iter):
                    # print(f"start clean iteration {J+1} of max {self.n_iter}")
                    if self.r_diag:
                        powPhi = (Phi[:num]**2-autopow).sum(0).clip(min=0)
                    else:
                        powPhi = (Phi[:num]**2).sum(0)
                    imax = argmax(powPhi)
                    t_float = d_interp2[imax] + d_index[imax] + n_index
                    t_ind = t_float.astype(int64)
                    for m in range(numMics): 
                        p_res[t_ind[:num+1,m],m] -= self.damp*interp(t_ind[:num+1,m],
                                                                t_float[:num,m],
                                                                    Phi[:num,imax]*self.r0[imax]/self.rm[imax,m],
                                                                    )
                    nextPhi, nextAutopow = self.delay_and_sum(num,p_res,d_interp2,d_index,amp)
                    if self.r_diag:
                        pownextPhi = (nextPhi[:num]**2-nextAutopow).sum(0).clip(min=0)
                    else:
                        pownextPhi = (nextPhi[:num]**2).sum(0)
                    # print(f"total signal power: {powPhi.sum()}")
                    if pownextPhi.sum() < powPhi.sum(): # stopping criterion
                        Gamma[:num,imax] += self.damp*Phi[:num,imax]
                        Gamma_autopow[:num,imax] = autopow[:num,imax].copy()
                        Phi=nextPhi
                        autopow=nextAutopow
                        # print(f"clean max: {L_p((Gamma**2).sum(0)/num).max()} dB")
                        J += 1
                    else:
                        break
                if 'Sq' not in self.__class__.__name__:
                    yield Gamma[:num]
                elif self.r_diag:                 
                    yield Gamma[:num]**2 - (self.damp**2)*Gamma_autopow[:num]
                else:
                    yield Gamma[:num]**2
            self.bufferIndex += num
            try:
                next(fill_buffer_generator)
            except: 
                pass

    def delay_and_sum(self,num,p_res,d_interp2,d_index,amp): 
        ''' standard delay-and-sum method ''' 
        result = empty((num, self.grid.size), dtype=float) # output array
        autopow = empty((num, self.grid.size), dtype=float) # output array
        _delayandsum4(p_res, d_index, d_interp2, amp, result, autopow)
        return result, autopow          
            

class BeamformerTimeSq( BeamformerTime ):
    """
    Provides a time domain beamformer with time-dependend
    power signal output and possible autopower removal
    for a spatially fixed grid.
    """
    
    #: Boolean flag, if 'True' (default), the main diagonal is removed before beamforming.
    r_diag = Bool(True, 
        desc="removal of diagonal")

    # internal identifier
    digest = Property( 
        depends_on = ['_steer_obj.digest', 'source.digest', 'r_diag', \
                      'weights', '__class__'], 
        )

    @cached_property
    def _get_digest( self ):
        return digest(self)
        

class BeamformerTimeTraj( BeamformerTime ):
    """
    Provides a basic time domain beamformer with time signal output
    for a grid moving along a trajectory.
    """


    #: :class:`~acoular.trajectory.Trajectory` or derived object.
    #: Start time is assumed to be the same as for the samples.
    trajectory = Trait(Trajectory, 
        desc="trajectory of the grid center")

    #: Reference vector, perpendicular to the y-axis of moving grid.
    rvec = CArray( dtype=float, shape=(3, ), value=array((0, 0, 0)), 
        desc="reference vector")
    
    #: Considering of convective amplification in beamforming formula.
    conv_amp = Bool(False, 
        desc="determines if convective amplification of source is considered")

    #: Floating point and integer precision
    precision = Trait(64, [32,64], 
        desc="numeric precision")

    # internal identifier
    digest = Property( 
        depends_on = ['_steer_obj.digest', 'source.digest', 'weights', 'precision',\
                      'rvec','conv_amp','trajectory.digest', '__class__'], 
        )

    @cached_property
    def _get_digest( self ):
        return digest(self)
        
    def get_moving_gpos(self):
        """
        Python generator that yields the moving grid coordinates samplewise 
        """
        def cross(a, b):
            """ cross product for fast computation
                because numpy.cross is ultra slow in this case
            """
            return array([a[1]*b[2] - a[2]*b[1],
                        a[2]*b[0] - a[0]*b[2],
                        a[0]*b[1] - a[1]*b[0]])

        start_t = 0.0
        gpos = self.grid.pos()
        trajg = self.trajectory.traj( start_t, delta_t=1/self.source.sample_freq)
        trajg1 = self.trajectory.traj( start_t, delta_t=1/self.source.sample_freq, 
                                  der=1)
        rflag = (self.rvec == 0).all() #flag translation vs. rotation
        if rflag:
            for g in trajg:
                # grid is only translated, not rotated
                tpos = gpos + array(g)[:, newaxis]
                yield tpos
        else:
            for (g,g1) in zip(trajg,trajg1):
                # grid is both translated and rotated
                loc = array(g) #translation array([0., 0.4, 1.])
                dx = array(g1) #direction vector (new x-axis)
                dy = cross(self.rvec, dx) # new y-axis
                dz = cross(dx, dy) # new z-axis
                RM = array((dx, dy, dz)).T # rotation matrix
                RM /= sqrt((RM*RM).sum(0)) # column normalized
                tpos = dot(RM, gpos)+loc[:, newaxis] # rotation+translation
#                print(loc[:])
                yield tpos

    def get_macostheta(self,g1,tpos,rm):
        vvec = array(g1) # velocity vector
        ma = norm(vvec)/self.steer.env.c # machnumber
        fdv = (vvec/sqrt((vvec*vvec).sum()))[:, newaxis] # unit vecor velocity
        mpos = self.steer.mics.mpos[:, newaxis, :]
        rmv = tpos[:, :, newaxis]-mpos
        return (ma*sum(rmv.reshape((3, -1))*fdv, 0)\
                                  /rm.reshape(-1)).reshape(rm.shape)

    def get_r0( self, tpos ):
        if isscalar(self.steer.ref) and self.steer.ref > 0:
            return self.steer.ref#full((self.steer.grid.size,), self.steer.ref)
        else:
            return self.env._r(tpos)

    def increase_buffer( self, num ): 
        ar = zeros((num,self.steer.mics.num_mics), dtype=self.buffer.dtype)
        self.buffer = concatenate((ar,self.buffer), axis=0)
        self.bufferIndex += num

    def result( self, num=2048 ):
        """
        Python generator that yields the deconvolved output block-wise. 
        
        Parameters
        ----------
        num : integer, defaults to 2048
            This parameter defines the size of the blocks to be yielded
            (i.e. the number of samples per block).
        
        Returns
        -------
        Samples in blocks of shape  \
        (num, :attr:`~BeamformerTime.numchannels`). 
            :attr:`~BeamformerTime.numchannels` is usually very \
            large (number of grid points).
            The last block may be shorter than num. \
            The output starts for signals that were emitted 
            from the grid at `t=0`.
        """
        # initialize values
        if self.precision==64:
            fdtype = float64
            idtype = int64
        else:
            fdtype = float32
            idtype = int32
        w = self._get_weights()
        c = self.steer.env.c/self.source.sample_freq
        numMics = self.steer.mics.num_mics
        mpos = self.steer.mics.mpos.astype(fdtype)
        m_index = arange(numMics, dtype=idtype)
        n_index = arange(num,dtype=idtype)[:,newaxis]
        blockrm = empty((num,self.grid.size,numMics),dtype=fdtype)
        blockrmconv = empty((num,self.grid.size,numMics),dtype=fdtype)
        amp = empty((num,self.grid.size,numMics),dtype=fdtype)
        #delays = empty((num,self.grid.size,numMics),dtype=fdtype)
        d_index = empty((num,self.grid.size,numMics),dtype=idtype)
        d_interp2 = empty((num,self.grid.size,numMics),dtype=fdtype)
        blockr0 = empty((num,self.grid.size),dtype=fdtype)
        self.buffer = zeros((2*num,numMics), dtype=fdtype)
        self.bufferIndex = self.buffer.shape[0] 
        movgpos = self.get_moving_gpos() # create moving grid pos generator
        movgspeed = self.trajectory.traj(0.0, delta_t=1/self.source.sample_freq, 
              der=1)
        fill_buffer_generator = self._fill_buffer(num)
        for i in range(2): 
            next(fill_buffer_generator)
        
        # preliminary implementation of different steering vectors
        steer_func = {'classic' : _steer_I,
                'inverse' : _steer_II,
                'true level' : _steer_III,
                'true location' : _steer_IV
                }[self.steer.steer_type]

        # start processing
        flag = True
        dflag = True # data is available 
        while flag:
            for i in range(num):
                tpos = next(movgpos).astype(fdtype)
                rm = self.steer.env._r( tpos, mpos )#.astype(fdtype) 
                blockr0[i,:] = self.get_r0(tpos)
                blockrm[i,:,:] = rm
                if self.conv_amp:
                    ht = next(movgspeed)
                    blockrmconv[i,:,:] = rm*(1-self.get_macostheta(ht,tpos,rm))**2
            if self.conv_amp:
                steer_func(blockrmconv, blockr0, amp)
            else:
                steer_func(blockrm, blockr0, amp)
            _delays(blockrm, c, d_interp2, d_index)
            #_modf(delays, d_interp2, d_index)
            maxdelay = (d_index.max((1,2)) + arange(0,num)).max()+2 # + because of interpolation
            # increase buffer size because of greater delays
            while maxdelay > self.buffer.shape[0] and dflag:
                self.increase_buffer(num)
                try:
                    next(fill_buffer_generator)
                except:
                    dflag = False
            samplesleft = self.buffer.shape[0]-self.bufferIndex
            # last block may be shorter
            if samplesleft-maxdelay <= 0:
                num = sum((d_index.max((1,2))+1+arange(0,num)) < samplesleft)
                n_index = arange(num,dtype=idtype)[:,newaxis]
                flag=False
            # init step
            p_res = self.buffer[self.bufferIndex:self.bufferIndex+maxdelay,:].copy()
            Phi, autopow = self.delay_and_sum(num,p_res,d_interp2,d_index,amp)
            if 'Cleant' not in self.__class__.__name__:
                if 'Sq' not in self.__class__.__name__:
                    yield Phi[:num]
                elif self.r_diag:                 
                    yield (Phi[:num]**2 - autopow[:num]).clip(min=0)
                else:
                    yield Phi[:num]**2
            else:
                # choose correct distance
                if self.conv_amp: 
                    blockrm1 = blockrmconv
                else:
                    blockrm1 = blockrm
                Gamma = zeros(Phi.shape,dtype=fdtype)
                Gamma_autopow = zeros(Phi.shape,dtype=fdtype)
                J = 0
                t_ind = arange(p_res.shape[0],dtype=idtype)
                # deconvolution 
                while (J < self.n_iter):
                    # print(f"start clean iteration {J+1} of max {self.n_iter}")
                    if self.r_diag:
                        powPhi = (Phi[:num]*Phi[:num]-autopow).sum(0).clip(min=0)
                    else:
                        powPhi = (Phi[:num]*Phi[:num]).sum(0)
                    # find index of max power focus point
                    imax = argmax(powPhi)
                    # find backward delays
                    t_float = (d_interp2[:num,imax,m_index] + \
                               d_index[:num,imax,m_index] + \
                               n_index).astype(fdtype)
                    # determine max/min delays in sample units
                    # + 2 because we do not want to extrapolate behind the last sample
                    ind_max = t_float.max(0).astype(idtype)+2 
                    ind_min = t_float.min(0).astype(idtype)
                    # store time history at max power focus point
                    h = Phi[:num,imax]*blockr0[:num,imax]
                    for m in range(numMics):
                        # subtract interpolated time history from microphone signals
                        p_res[ind_min[m]:ind_max[m],m] -= self.damp*interp(
                            t_ind[ind_min[m]:ind_max[m]], 
                            t_float[:num,m],
                            h/blockrm1[:num,imax,m],
                                )
                    nextPhi, nextAutopow = self.delay_and_sum(num,p_res,d_interp2,d_index,amp)
                    if self.r_diag:
                        pownextPhi = (nextPhi[:num]*nextPhi[:num]-nextAutopow).sum(0).clip(min=0)
                    else:
                        pownextPhi = (nextPhi[:num]*nextPhi[:num]).sum(0)                 
                    # print(f"total signal power: {powPhi.sum()}")
                    if pownextPhi.sum() < powPhi.sum(): # stopping criterion
                        Gamma[:num,imax] += self.damp*Phi[:num,imax]
                        Gamma_autopow[:num,imax] = autopow[:num,imax].copy()
                        Phi=nextPhi
                        autopow=nextAutopow
                        # print(f"clean max: {L_p((Gamma**2).sum(0)/num).max()} dB")
                        J += 1
                    else:
                        break
                if 'Sq' not in self.__class__.__name__:
                    yield Gamma[:num]
                elif self.r_diag: 
                    yield Gamma[:num]**2 - (self.damp**2)*Gamma_autopow[:num]
                else:
                    yield Gamma[:num]**2
            self.bufferIndex += num
            try:
                next(fill_buffer_generator)
            except: 
                dflag = False
                pass

    def delay_and_sum(self,num,p_res,d_interp2,d_index,amp): 
        ''' standard delay-and-sum method ''' 
        if self.precision==64:
            fdtype = float64
        else:
            fdtype = float32
        result = empty((num, self.grid.size), dtype=fdtype) # output array
        autopow = empty((num, self.grid.size), dtype=fdtype) # output array
        _delayandsum5(p_res, d_index, d_interp2, amp, result, autopow)
        return result, autopow          
  
        
class BeamformerTimeSqTraj( BeamformerTimeSq, BeamformerTimeTraj ):
    """
    Provides a time domain beamformer with time-dependent
    power signal output and possible autopower removal
    for a grid moving along a trajectory.
    """
    
    # internal identifier
    digest = Property( 
        depends_on = ['_steer_obj.digest', 'source.digest', 'r_diag', 'weights', 'precision',\
                      'rvec','conv_amp','trajectory.digest', '__class__'], 
        )

    @cached_property
    def _get_digest( self ):
        return digest(self)
                               

class BeamformerCleant( BeamformerTime ):
    """
    CLEANT deconvolution method, see :ref:`Cousson et al., 2019<Cousson2019>`.
    
    An implementation of the CLEAN method in time domain. This class can only 
    be used for static sources.
    """

    #: Boolean flag, always False
    r_diag = Enum(False, 
                  desc="False, as we do not remove autopower in this beamformer")

    #: iteration damping factor also referred as loop gain in Cousson et al. 
    #: defaults to 0.6
    damp = Range(0.01, 1.0, 0.6, 
        desc="damping factor (loop gain)")

    #: max number of iterations
    n_iter = Int(100, 
        desc="maximum number of iterations")
     
    # internal identifier
    digest = Property( 
        depends_on = ['_steer_obj.digest', 'source.digest', 'weights',  \
                      '__class__','damp','n_iter'],
        )

    @cached_property
    def _get_digest( self ):
        return digest(self)


class BeamformerCleantSq( BeamformerCleant ):
    """
    CLEANT deconvolution method, see :ref:`Cousson et al., 2019<Cousson2019>`
    with optional removal of autocorrelation.
    
    An implementation of the CLEAN method in time domain. This class can only 
    be used for static sources.
    """

    #: Boolean flag, if 'True' (default), the main diagonal is removed before beamforming.
    r_diag = Bool(True, 
        desc="removal of diagonal")

    # internal identifier
    digest = Property( 
        depends_on = ['_steer_obj.digest', 'source.digest', 'weights',  \
                      '__class__','damp','n_iter','r_diag'],
        )

    @cached_property
    def _get_digest( self ):
        return digest(self)
    
    

class BeamformerCleantTraj( BeamformerCleant, BeamformerTimeTraj ):
    """
    CLEANT deconvolution method, see :ref:`Cousson et al., 2019<Cousson2019>`.
    
    An implementation of the CLEAN method in time domain for moving sources
    with known trajectory. 
    """
    #: Floating point and integer precision
    precision = Trait(32, [32,64], 
        desc="numeric precision")

    # internal identifier
    digest = Property( 
        depends_on = ['_steer_obj.digest', 'source.digest', 'weights', 'precision', \
                      '__class__','damp','n_iter', 'rvec','conv_amp',
                      'trajectory.digest'],
        )

    @cached_property
    def _get_digest( self ):
        return digest(self)
    

class BeamformerCleantSqTraj( BeamformerCleantTraj, BeamformerTimeSq ):
    """
    CLEANT deconvolution method, see :ref:`Cousson et al., 2019<Cousson2019>`
    with optional removal of autocorrelation.
    
    An implementation of the CLEAN method in time domain for moving sources
    with known trajectory. 
    """

    #: Boolean flag, if 'True' (default), the main diagonal is removed before beamforming.
    r_diag = Bool(True, 
        desc="removal of diagonal")

    # internal identifier
    digest = Property( 
        depends_on = ['_steer_obj.digest', 'source.digest', 'weights', 'precision', \
                      '__class__','damp','n_iter', 'rvec','conv_amp',
                      'trajectory.digest','r_diag'],
        )

    @cached_property
    def _get_digest( self ):
        return digest(self)
    

class IntegratorSectorTime( TimeInOut ):
    """
    Provides an Integrator in the time domain.
    """

    #: :class:`~acoular.grids.RectGrid` object that provides the grid locations.
    grid = Trait(RectGrid, 
        desc="beamforming grid")
        
    #: List of sectors in grid
    sectors = List()

    #: Clipping, in Dezibel relative to maximum (negative values)
    clip = Float(-350.0)

    #: Number of channels in output (= number of sectors).
    numchannels = Property( depends_on = ['sectors', ])

    # internal identifier
    digest = Property( 
        depends_on = ['sectors', 'clip', 'grid.digest', 'source.digest', \
        '__class__'], 
        )

    @cached_property
    def _get_digest( self ):
        return digest(self)
        
    @cached_property
    def _get_numchannels ( self ):
        return len(self.sectors)

    def result( self, num=1 ):
        """
        Python generator that yields the source output integrated over the given 
        sectors, block-wise.
        
        Parameters
        ----------
        num : integer, defaults to 1
            This parameter defines the size of the blocks to be yielded
            (i.e. the number of samples per block).
        
        Returns
        -------
        Samples in blocks of shape (num, :attr:`numchannels`). 
        :attr:`numchannels` is the number of sectors.
        The last block may be shorter than num.
        """
        inds = [self.grid.indices(*sector) for sector in self.sectors]
        gshape = self.grid.shape
        o = empty((num, self.numchannels), dtype=float) # output array
        for r in self.source.result(num):
            ns = r.shape[0]
            mapshape = (ns,) + gshape
            rmax = r.max()
            rmin = rmax * 10**(self.clip/10.0)
            r = where(r>rmin, r, 0.0)
            i = 0
            for ind in inds:
                h = r[:].reshape(mapshape)[ (s_[:],) + ind ]
                o[:ns, i] = h.reshape(h.shape[0], -1).sum(axis=1)
                i += 1
            yield o[:ns]


#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Checking various implementations of the damas solver


Versions used in this script:
numba=0.34.0
python=2.7.13
"""
import time as tm
import threading
import gc

import numpy as np
import numba as nb

import sharedFunctions as shFncs
from beamformer import gseidel1  # The benchmark (created with scipy.weave)
#from beamformer_withoutMP import r_beamfull_inverse_OhneMP  # also created with scipy.weave, but WITHOUT using multiple cores via OpenMP

@nb.njit(nb.float32[:](nb.float32[:,:], nb.float32[:], nb.float32[:], nb.int64), cache=True)
def njit_pure(A, dirtyMap, damasSolution, nIterations):
    nGridPoints = len(dirtyMap)
    for cntIter in xrange(nIterations):
        for cntGrid in xrange(nGridPoints):
            solHelp = np.float32(0)
            for cntGridHelp in xrange(cntGrid):
                solHelp += A[cntGrid, cntGridHelp] * damasSolution[cntGridHelp]
            for cntGridHelp in xrange(cntGrid + 1, nGridPoints):
                solHelp += A[cntGrid, cntGridHelp] * damasSolution[cntGridHelp]
            solHelp = dirtyMap[cntGrid] - solHelp
            if solHelp > 0.0:
                damasSolution[cntGrid] = solHelp
            else:
                damasSolution[cntGrid] = 0.0
    return damasSolution

@nb.njit(nb.float32[:](nb.float32[:,:], nb.float32[:], nb.float32[:], nb.int64), parallel=True)
def njit_parallel(A, dirtyMap, damasSolution, nIterations):
    nGridPoints = len(dirtyMap)
    for cntIter in xrange(nIterations):
        for cntGrid in xrange(nGridPoints):
            solHelp = np.float32(0)
            for cntGridHelp in xrange(cntGrid):
                solHelp += A[cntGrid, cntGridHelp] * damasSolution[cntGridHelp]
            for cntGridHelp in xrange(cntGrid + 1, nGridPoints):
                solHelp += A[cntGrid, cntGridHelp] * damasSolution[cntGridHelp]
            solHelp = dirtyMap[cntGrid] - solHelp
            if solHelp > 0.0:
                damasSolution[cntGrid] = solHelp
            else:
                damasSolution[cntGrid] = 0.0
    return damasSolution

@nb.guvectorize([(nb.float32[:,:], nb.float32[:], nb.int64, nb.float32[:])], '(g,g),(g),()->(g)', cache=True)
def guvectorizeOverFreqs(A, dirtyMap, nIterations, damasSolution):
    nGridPoints = len(dirtyMap)
    for cntIter in xrange(nIterations):
        for cntGrid in xrange(nGridPoints):
            solHelp = np.float32(0)
            for cntGridHelp in xrange(cntGrid):
                solHelp += A[cntGrid, cntGridHelp] * damasSolution[cntGridHelp]
            for cntGridHelp in xrange(cntGrid + 1, nGridPoints):
                solHelp += A[cntGrid, cntGridHelp] * damasSolution[cntGridHelp]
            solHelp = dirtyMap[cntGrid] - solHelp
            if solHelp > 0.0:
                damasSolution[cntGrid] = solHelp
            else:
                damasSolution[cntGrid] = 0.0


def njit_multiThreading(A, dirtyMap, damasSolution, nIterations):
    nGridPoints = len(dirtyMap)
    for cntIter in xrange(nIterations):
        for cntGrid in xrange(nGridPoints):
            sumHelp = np.zeros((2), np.float32)
            threadLowerSum = threading.Thread(target=njit_coreSum, args=(A[cntGrid, :], damasSolution, 0, cntGrid, sumHelp, 0))
            threadUpperSum = threading.Thread(target=njit_coreSum, args=(A[cntGrid, :], damasSolution, cntGrid+1, nGridPoints, sumHelp, 1))
            threadLowerSum.start()
            threadUpperSum.start()
            threadLowerSum.join()
            threadUpperSum.join()
            solHelp = np.float32(dirtyMap[cntGrid] - (sumHelp[0] + sumHelp[1]))
            if solHelp > 0.0:
                damasSolution[cntGrid] = solHelp
            else:
                damasSolution[cntGrid] = 0.0
    return damasSolution
@nb.njit(nb.void(nb.float32[:], nb.float32[:], nb.int64, nb.int64, nb.float32[:], nb.int64), cache=True, nogil=True)
def njit_coreSum(A, damasSolution, start, stop, result, indRes):
    for cntGridHelp in xrange(start, stop):
        result[indRes] += A[cntGridHelp] * damasSolution[cntGridHelp]

#%%  MAIN
listOfMics = [0] # not really needed
listGridPoints = [500, 1000, 2000, 5000, 10000]  # Standard value: 12000  # The number of gridpoints doesn't seeme to have to great of an influence
nTrials = 10
nIterations = 10
listOfNFreqs = [20]

funcsToTrial = [njit_pure, njit_parallel, guvectorizeOverFreqs, njit_multiThreading, gseidel1]
for nMics in listOfMics:
    for nGridPoints in listGridPoints:
        for nFreqs in listOfNFreqs:
            # Init
            print(10*'-' + 'New Test configuration: nMics=%s, nGridPoints=%s, nFreqs=%s' %(nMics, nGridPoints, nFreqs) + 10*'-')
            print(10*'-' + 'Creation of inputInputs' + 10*'-')
            
            A = np.float32(np.random.rand(1, nGridPoints, nGridPoints))  #A = np.float32(np.ones((1, nGridPoints, nGridPoints)))
#            for cntFreqs in range(nFreqs):
#                A[cntFreqs, :, :] += A[cntFreqs, :, :].T.conj()
            dirtyMap = np.float32(np.random.rand(1, nGridPoints))  #dirtyMap = np.float32(np.ones((1, nGridPoints)))
            damasSolution = np.zeros((nGridPoints), np.float32)
            
            # create nFreqs times the same matrix for comparing reasons
            A = np.tile(A, [nFreqs, 1, 1])
            dirtyMap = np.tile(dirtyMap, [nFreqs, 1])
            
            nameOfFuncsToTrial = map(lambda x: x.__name__, funcsToTrial)
            nameOfFuncsForError = [funcName for funcName in nameOfFuncsToTrial if funcName != 'gseidel1']
            maxRelativeDeviation = np.zeros((len(funcsToTrial), nTrials))
            maxAbsoluteDeviation = np.zeros((len(funcsToTrial), nTrials))
            timeConsumption = [[] for _ in range(len(funcsToTrial))]
            indOfBaselineFnc = nameOfFuncsToTrial.index('gseidel1')
            
            print(10*'-' + 'Onetime calculation of error reference' + 10*'-')
            gseidel1(A[0,:,:], dirtyMap[0,:], damasSolution, nIterations)
            resultReference = damasSolution  # For relative/absolute error
            gc.collect()
            
            # Testing
            print(10*'-' + 'Testing of functions' + 10*'-')
            cntFunc = 0
            for func in funcsToTrial:
                print(func.__name__)
                for cntTrials in xrange(nTrials):
                    damasSolution = np.zeros((1, nGridPoints), np.float32)
                    damasSolution = np.tile(damasSolution, [nFreqs, 1])
                    if func.__name__ == 'guvectorizeOverFreqs':
                        t0 = tm.time()
                        func(A, dirtyMap, nIterations, damasSolution)
                        t1 = tm.time()
                        result = damasSolution[0, :]
                    elif func.__name__ == 'gseidel1':
                        t0 = tm.time()
                        for cntFreqsHelp in xrange(nFreqs):
                            func(A[cntFreqsHelp,:,:], dirtyMap[cntFreqsHelp,:], damasSolution[cntFreqsHelp,:], nIterations)
                        t1 = tm.time()
                        result = damasSolution[0, :]
                    else:
                        t0 = tm.time()
                        for cntFreqsHelp in xrange(nFreqs):
                            output = func(A[cntFreqsHelp,:,:], dirtyMap[cntFreqsHelp,:], damasSolution[cntFreqsHelp,:], nIterations)
                        t1 = tm.time()
                        result = output
                    timeConsumption[cntFunc].append(t1 - t0)
                    relativeDiffBetweenNewCodeAndRef = (result - resultReference) / (result + resultReference) * 2  # error in relation to the resulting value
                    maxRelativeDeviation[cntFunc, cntTrials] = np.amax(abs(relativeDiffBetweenNewCodeAndRef), axis=0)  # relative error in inf-norm
                    maxAbsoluteDeviation[cntFunc, cntTrials] = np.amax(abs(result - resultReference), axis=0)  # absolute error in inf-norm
                cntFunc += 1
            factorTimeConsump = [np.mean(timeConsumption[cnt]) for cnt in range(0, len(funcsToTrial))] \
                                / np.mean(timeConsumption[indOfBaselineFnc])
            
            # Save the current test-config as .sav
            helpString = 'The order of the variables is: \n nameOfFuncsToTrial \n maxRelativeDeviation'\
                '\n timeConsumption [nFuncs, nTrials] \n nMics \n nGridPoints \n nFreqs '\
                '\n Factor of time consumption (in relation to the original .cpp) \n maxAbsoluteDeviation \n nThreadsGlobal'
            saveTupel = (helpString, nameOfFuncsToTrial, maxRelativeDeviation, timeConsumption, 
                         nMics, nGridPoints, nFreqs, factorTimeConsump, maxAbsoluteDeviation, 0)
            stringParameters = 'OvernightTestcasesBeamformer_nMics%s_nGridPoints%s_nFreqs%s_nTrials%s' %(nMics, nGridPoints, nFreqs, nTrials)
            
#            stringSaveName = 'Peter'
            stringSaveName = 'Sicherung_DurchgelaufeneTests/damasSolver/' + stringParameters
#            stringSaveName = 'Sicherung_DurchgelaufeneTests/Beamformer/AllImportantMethods/' + stringParameters
#            stringSaveName = 'Sicherung_DurchgelaufeneTests/Beamformer/EinflussGridpoints/AMDFX6100/' + stringParameters
#            stringSaveName = 'Sicherung_DurchgelaufeneTests/Beamformer/JitPrange/' + stringParameters
#            stringSaveName = 'Sicherung_DurchgelaufeneTests/Beamformer/Multithreading_02Threads/' + stringParameters

            shFncs.savingTimeConsumption(stringSaveName, saveTupel)  # saving as "stringSaveName.sav"
            
#            shFncs.plottingOfOvernightTestcasesBeamformer(stringSaveName + '.sav')  # plot of the current test-config



#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Thu Jul 20 14:06:46 2017

@author: tomgensch

This script was for getting to know the mechanisms of the new prange function
of jit (numba version 0.34). As it turns out the automatic parallelization is
as faulty as in cython (same errors).
"""

import numpy as np
from numba import njit, float64, complex128, prange

nMics = 4
nGridPoints = 5
nFreqs = 1

csm = np.random.rand(nFreqs, nMics, nMics) + 1j*np.random.rand(nFreqs, nMics, nMics)
for cntFreqs in range(nFreqs):
    csm[cntFreqs, :, :] += csm[cntFreqs, :, :].T.conj()  # zu Hermitischer Matrix machen
r0 = np.random.rand(nGridPoints) #abstand aufpunkte-arraymittelpunkt
rm = np.random.rand(nGridPoints, nMics) #abstand aufpunkte-arraymikrofone
kj = np.zeros(nFreqs) + 1j*np.random.rand(nFreqs) 

@njit(float64[:,:](complex128[:,:,:], float64[:], float64[:,:], complex128[:]), parallel=True)
def loops_NumbaJit_parallelSlow(csm, r0, rm, kj):
    nFreqs = csm.shape[0]
    nGridPoints = len(r0)
    nMics = csm.shape[1]
    beamformOutput = np.zeros((nFreqs, nGridPoints), np.float64)
    
    for cntFreqs in xrange(nFreqs):
        kjj = kj[cntFreqs].imag
        for cntGrid in prange(nGridPoints):
            r01 = r0[cntGrid]
            rs = r01 ** 2
            
            temp1 = 0.0
            for cntMics in xrange(nMics): 
                temp2 = 0.0
                rm1 = rm[cntGrid, cntMics]
                temp3 = np.float32(kjj * (rm1 - r01))
                steerVec = (np.cos(temp3) - 1j * np.sin(temp3)) * rm1
                
                for cntMics2 in xrange(cntMics):
                    rm1 = rm[cntGrid, cntMics2]
                    temp3 = np.float32(kjj * (rm1 - r01))
                    steerVec1 = (np.cos(temp3) - 1j * np.sin(temp3)) * rm1
                    temp2 += csm[cntFreqs, cntMics2, cntMics] * steerVec1
                temp1 += 2 * (temp2 * steerVec.conjugate()).real
                temp1 += (csm[cntFreqs, cntMics, cntMics] * steerVec.conjugate() * steerVec).real
            
            beamformOutput[cntFreqs, cntGrid] = (temp1 / rs).real
    return beamformOutput

@njit(float64[:,:](complex128[:,:,:], float64[:], float64[:,:], complex128[:]), parallel=True)
def loops_NumbaJit_parallelFast(csm, r0, rm, kj):
    """ This method implements the prange over the Gridpoints, which is a direct
    implementation of the currently used c++ methods created with scipy.wave.
    
    Very strange: Just like with Cython, this implementation (prange over Gridpoints)
    produces wrong results. If one doesn't parallelize -> everything is good 
    (just like with Cython). Maybe Cython and Numba.jit use the same interpreter 
    to generate OpenMP-parallelizable code.
    
    BUT: If one uncomments the 'steerVec' declaration in the prange-loop over the
    gridpoints an error occurs. After commenting the line again and executing
    the script once more, THE BEAMFORMER-RESULTS ARE CORRECT (for repeated tries). 
    Funny enough the method is now twice as slow in comparison to the 
    'wrong version' (before invoking the error).
    """
    # init
    nFreqs = csm.shape[0]
    nGridPoints = len(r0)
    nMics = csm.shape[1]
    beamformOutput = np.zeros((nFreqs, nGridPoints), np.float64)
    steerVec = np.zeros((nMics), np.complex128)
    
    for cntFreqs in xrange(nFreqs):
        kjj = kj[cntFreqs].imag
        for cntGrid in prange(nGridPoints):
#            steerVec = np.zeros((nMics), np.complex128)  # This is the line that has to be uncommented (see this methods documentation comment)
            rs = 0
            r01 = r0[cntGrid]
            
            for cntMics in xrange(nMics):
                rm1 = rm[cntGrid, cntMics]
                rs += 1.0 / (rm1**2)
                temp3 = np.float32(kjj * (rm1 - r01))
                steerVec[cntMics] = (np.cos(temp3) - 1j * np.sin(temp3)) * rm1
            rs = r01 ** 2
            
            temp1 = 0.0
            for cntMics in xrange(nMics):
                temp2 = 0.0
                for cntMics2 in xrange(cntMics):
                    temp2 = temp2 + csm[cntFreqs, cntMics2, cntMics] * steerVec[cntMics2]
                temp1 = temp1 + 2 * (temp2 * steerVec[cntMics].conjugate()).real
                temp1 = temp1 + (csm[cntFreqs, cntMics, cntMics] * np.conjugate(steerVec[cntMics]) * steerVec[cntMics]).real
            
            beamformOutput[cntFreqs, cntGrid] = (temp1 / rs).real
    return beamformOutput

beamformOutputFast = loops_NumbaJit_parallelFast(csm, r0, rm, kj)
beamformOutputSlow = loops_NumbaJit_parallelSlow(csm, r0, rm, kj)

#%% Referenz
beamformOutputRef = np.zeros((nFreqs, nGridPoints), np.float64)
e1 = np.zeros((nMics), np.complex128)

for cntFreqs in xrange(nFreqs):
    kjj = kj[cntFreqs].imag
    for cntGrid in xrange(nGridPoints):
        rs = 0
        r01 = r0[cntGrid]
        
        for cntMics in xrange(nMics):
            rm1 = rm[cntGrid, cntMics]
            rs += 1.0 / (rm1**2)
            temp3 = np.float32(kjj * (rm1 - r01)) 
            e1[cntMics] = (np.cos(temp3) - 1j * np.sin(temp3)) * rm1
        rs = r01 ** 2
        
        temp1 = 0.0
        for cntMics in xrange(nMics):
            temp2 = 0.0
            for cntMics2 in xrange(cntMics):
                temp2 += csm[cntFreqs, cntMics2, cntMics] * e1[cntMics2]
            temp1 += 2 * (temp2 * e1[cntMics].conjugate()).real
            temp1 += (csm[cntFreqs, cntMics, cntMics] * np.conjugate(e1[cntMics]) * e1[cntMics]).real
        
        beamformOutputRef[cntFreqs, cntGrid] = (temp1 / rs).real

diffFast = np.amax(abs(beamformOutputFast - beamformOutputRef), axis=0)
diffFast= np.amax(abs(diffFast), axis=0)

diffSlow = np.amax(abs(beamformOutputSlow - beamformOutputRef), axis=0)
diffSlow= np.amax(abs(diffSlow), axis=0)

print(diffFast)
print(diffSlow)


# -*- coding: utf-8 -*-
"""
This script is a help for checking if the new NUMBA functions are correctly 
integrated into acoular.
One has to make a savefile (see 'all_bfWeave.sav') for both, the old acoular 
version an the new one. In section '#%% Compare Weave vs Numba' both versions 
are  compared.

This script uses essentially 'example3.py', so therefor 'example_data.h5' and 
'example_calib.xml' are needed.


Copyright (c) 2006-2015 The Acoular developers.
All rights reserved.
"""

# imports from acoular
import acoular
from acoular import L_p, TimeSamples, Calib, MicGeom, EigSpectra,\
RectGrid3D, BeamformerBase, BeamformerFunctional, BeamformerEig, BeamformerOrth, \
BeamformerCleansc, BeamformerCapon, BeamformerMusic, BeamformerCMF, PointSpreadFunction, BeamformerClean, BeamformerDamas

# other imports
from os import path
#from mayavi import mlab
from numpy import amax
#from cPickle import dump, load
from pickle import dump, load

# see example3
t = TimeSamples(name='example_data.h5')
cal = Calib(from_file='example_calib.xml')
m = MicGeom(from_file=path.join(\
    path.split(acoular.__file__)[0], 'xml', 'array_56.xml'))
g = RectGrid3D(x_min=-0.6, x_max=-0.0, y_min=-0.3, y_max=0.3, \
    z_min=0.48, z_max=0.88, increment=0.1)
f = EigSpectra(time_data=t, window='Hanning', overlap='50%', block_size=128, ind_low=5, ind_high=15)
csm = f.csm[:]
eva = f.eva[:]
eve = f.eve[:]

#""" Creating the beamformers
bb1Rem = BeamformerBase(freq_data=f, grid=g, mpos=m, r_diag=True, c=346.04, steer='classic')
bb2Rem = BeamformerBase(freq_data=f, grid=g, mpos=m, r_diag=True, c=346.04, steer='inverse')
bb3Rem = BeamformerBase(freq_data=f, grid=g, mpos=m, r_diag=True, c=346.04, steer='true level')
bb4Rem = BeamformerBase(freq_data=f, grid=g, mpos=m, r_diag=True, c=346.04, steer='true location')
bb1Full = BeamformerBase(freq_data=f, grid=g, mpos=m, r_diag=False, c=346.04, steer='classic')
bb2Full = BeamformerBase(freq_data=f, grid=g, mpos=m, r_diag=False, c=346.04, steer='inverse')
bb3Full = BeamformerBase(freq_data=f, grid=g, mpos=m, r_diag=False, c=346.04, steer='true level')
bb4Full = BeamformerBase(freq_data=f, grid=g, mpos=m, r_diag=False, c=346.04, steer='true location')
Lbb1Rem = L_p(bb1Rem.synthetic(4000,1))
Lbb2Rem = L_p(bb2Rem.synthetic(4000,1))
Lbb3Rem = L_p(bb3Rem.synthetic(4000,1))
Lbb4Rem = L_p(bb4Rem.synthetic(4000,1))
Lbb1Full = L_p(bb1Full.synthetic(4000,1))
Lbb2Full = L_p(bb2Full.synthetic(4000,1))
Lbb3Full = L_p(bb3Full.synthetic(4000,1))
Lbb4Full = L_p(bb4Full.synthetic(4000,1))

bf1Rem = BeamformerFunctional(freq_data=f, grid=g, mpos=m, r_diag=True, c=346.04, steer='classic', gamma=3)
bf2Rem = BeamformerFunctional(freq_data=f, grid=g, mpos=m, r_diag=True, c=346.04, steer='inverse', gamma=3)
bf3Rem = BeamformerFunctional(freq_data=f, grid=g, mpos=m, r_diag=True, c=346.04, steer='true level', gamma=3)
bf4Rem = BeamformerFunctional(freq_data=f, grid=g, mpos=m, r_diag=True, c=346.04, steer='true location', gamma=3)
bf1Full = BeamformerFunctional(freq_data=f, grid=g, mpos=m, r_diag=False, c=346.04, steer='classic', gamma=3)
bf2Full = BeamformerFunctional(freq_data=f, grid=g, mpos=m, r_diag=False, c=346.04, steer='inverse', gamma=3)
bf3Full = BeamformerFunctional(freq_data=f, grid=g, mpos=m, r_diag=False, c=346.04, steer='true level', gamma=3)
bf4Full = BeamformerFunctional(freq_data=f, grid=g, mpos=m, r_diag=False, c=346.04, steer='true location', gamma=3)
Lbf1Rem = L_p(bf1Rem.synthetic(4000,1))
Lbf2Rem = L_p(bf2Rem.synthetic(4000,1))
Lbf3Rem = L_p(bf3Rem.synthetic(4000,1))
Lbf4Rem = L_p(bf4Rem.synthetic(4000,1))
Lbf1Full = L_p(bf1Full.synthetic(4000,1))
Lbf2Full = L_p(bf2Full.synthetic(4000,1))
Lbf3Full = L_p(bf3Full.synthetic(4000,1))
Lbf4Full = L_p(bf4Full.synthetic(4000,1))

bca1Full = BeamformerCapon(freq_data=f, grid=g, mpos=m, r_diag=False, c=346.04, steer='classic')
bca2Full = BeamformerCapon(freq_data=f, grid=g, mpos=m, r_diag=False, c=346.04, steer='inverse')
bca3Full = BeamformerCapon(freq_data=f, grid=g, mpos=m, r_diag=False, c=346.04, steer='true level')
bca4Full = BeamformerCapon(freq_data=f, grid=g, mpos=m, r_diag=False, c=346.04, steer='true location')
Lbca1Full = L_p(bca1Full.synthetic(4000,1))
Lbca2Full = L_p(bca2Full.synthetic(4000,1))
Lbca3Full = L_p(bca3Full.synthetic(4000,1))
Lbca4Full = L_p(bca4Full.synthetic(4000,1))

be1Rem = BeamformerEig(freq_data=f, grid=g, mpos=m, r_diag=True, c=346.04, steer='classic', n=12)
be2Rem = BeamformerEig(freq_data=f, grid=g, mpos=m, r_diag=True, c=346.04, steer='inverse', n=12)
be3Rem = BeamformerEig(freq_data=f, grid=g, mpos=m, r_diag=True, c=346.04, steer='true level', n=12)
be4Rem = BeamformerEig(freq_data=f, grid=g, mpos=m, r_diag=True, c=346.04, steer='true location', n=12)
be1Full = BeamformerEig(freq_data=f, grid=g, mpos=m, r_diag=False, c=346.04, steer='classic', n=12)
be2Full = BeamformerEig(freq_data=f, grid=g, mpos=m, r_diag=False, c=346.04, steer='inverse', n=12)
be3Full = BeamformerEig(freq_data=f, grid=g, mpos=m, r_diag=False, c=346.04, steer='true level', n=12)
be4Full = BeamformerEig(freq_data=f, grid=g, mpos=m, r_diag=False, c=346.04, steer='true location', n=12)
Lbe1Rem = L_p(be1Rem.synthetic(4000,1))
Lbe2Rem = L_p(be2Rem.synthetic(4000,1))
Lbe3Rem = L_p(be3Rem.synthetic(4000,1))
Lbe4Rem = L_p(be4Rem.synthetic(4000,1))
Lbe1Full = L_p(be1Full.synthetic(4000,1))
Lbe2Full = L_p(be2Full.synthetic(4000,1))
Lbe3Full = L_p(be3Full.synthetic(4000,1))
Lbe4Full = L_p(be4Full.synthetic(4000,1))

bm1Full = BeamformerMusic(freq_data=f, grid=g, mpos=m, r_diag=False, c=346.04, steer='classic', n=12)
bm2Full = BeamformerMusic(freq_data=f, grid=g, mpos=m, r_diag=False, c=346.04, steer='inverse', n=12)
bm3Full = BeamformerMusic(freq_data=f, grid=g, mpos=m, r_diag=False, c=346.04, steer='true level', n=12)
bm4Full = BeamformerMusic(freq_data=f, grid=g, mpos=m, r_diag=False, c=346.04, steer='true location', n=12)
Lbm1Full = L_p(bm1Full.synthetic(4000,1))
Lbm2Full = L_p(bm2Full.synthetic(4000,1))
Lbm3Full = L_p(bm3Full.synthetic(4000,1))
Lbm4Full = L_p(bm4Full.synthetic(4000,1))

bcsc1Rem = BeamformerCleansc(freq_data=f, grid=g, mpos=m, r_diag=True, c=346.04, steer='classic')
bcsc2Rem = BeamformerCleansc(freq_data=f, grid=g, mpos=m, r_diag=True, c=346.04, steer='inverse')
bcsc3Rem = BeamformerCleansc(freq_data=f, grid=g, mpos=m, r_diag=True, c=346.04, steer='true level')
bcsc4Rem = BeamformerCleansc(freq_data=f, grid=g, mpos=m, r_diag=True, c=346.04, steer='true location')
bcsc1Full = BeamformerCleansc(freq_data=f, grid=g, mpos=m, r_diag=False, c=346.04, steer='classic')
bcsc2Full = BeamformerCleansc(freq_data=f, grid=g, mpos=m, r_diag=False, c=346.04, steer='inverse')
bcsc3Full = BeamformerCleansc(freq_data=f, grid=g, mpos=m, r_diag=False, c=346.04, steer='true level')
bcsc4Full = BeamformerCleansc(freq_data=f, grid=g, mpos=m, r_diag=False, c=346.04, steer='true location')
Lbcsc1Rem = L_p(bcsc1Rem.synthetic(4000,1))
Lbcsc2Rem = L_p(bcsc2Rem.synthetic(4000,1))
Lbcsc3Rem = L_p(bcsc3Rem.synthetic(4000,1))
Lbcsc4Rem = L_p(bcsc4Rem.synthetic(4000,1))
Lbcsc1Full = L_p(bcsc1Full.synthetic(4000,1))
Lbcsc2Full = L_p(bcsc2Full.synthetic(4000,1))
Lbcsc3Full = L_p(bcsc3Full.synthetic(4000,1))
Lbcsc4Full = L_p(bcsc4Full.synthetic(4000,1))

bort1Rem = BeamformerOrth(beamformer=be1Rem, eva_list=list(range(4,8)))
bort2Rem = BeamformerOrth(beamformer=be2Rem, eva_list=list(range(4,8)))
bort3Rem = BeamformerOrth(beamformer=be3Rem, eva_list=list(range(4,8)))
bort4Rem = BeamformerOrth(beamformer=be4Rem, eva_list=list(range(4,8)))
bort1Full = BeamformerOrth(beamformer=be1Full, eva_list=list(range(4,8)))
bort2Full = BeamformerOrth(beamformer=be2Full, eva_list=list(range(4,8)))
bort3Full = BeamformerOrth(beamformer=be3Full, eva_list=list(range(4,8)))
bort4Full = BeamformerOrth(beamformer=be4Full, eva_list=list(range(4,8)))
Lbort1Rem = L_p(bort1Rem.synthetic(4000,1))
Lbort2Rem = L_p(bort2Rem.synthetic(4000,1))
Lbort3Rem = L_p(bort3Rem.synthetic(4000,1))
Lbort4Rem = L_p(bort4Rem.synthetic(4000,1))
Lbort1Full = L_p(bort1Full.synthetic(4000,1))
Lbort2Full = L_p(bort2Full.synthetic(4000,1))
Lbort3Full = L_p(bort3Full.synthetic(4000,1))
Lbort4Full = L_p(bort4Full.synthetic(4000,1))

bcmf1Rem = BeamformerCMF(freq_data=f, grid=g, mpos=m, r_diag=True, c=346.04, steer='classic')
bcmf2Rem = BeamformerCMF(freq_data=f, grid=g, mpos=m, r_diag=True, c=346.04, steer='inverse')
bcmf3Rem = BeamformerCMF(freq_data=f, grid=g, mpos=m, r_diag=True, c=346.04, steer='true level')
bcmf4Rem = BeamformerCMF(freq_data=f, grid=g, mpos=m, r_diag=True, c=346.04, steer='true location')
bcmf1Full = BeamformerCMF(freq_data=f, grid=g, mpos=m, r_diag=False, c=346.04, steer='classic')
bcmf2Full = BeamformerCMF(freq_data=f, grid=g, mpos=m, r_diag=False, c=346.04, steer='inverse')
bcmf3Full = BeamformerCMF(freq_data=f, grid=g, mpos=m, r_diag=False, c=346.04, steer='true level')
bcmf4Full = BeamformerCMF(freq_data=f, grid=g, mpos=m, r_diag=False, c=346.04, steer='true location')
Lbcmf1Rem = L_p(bcmf1Rem.synthetic(4000,1))
Lbcmf2Rem = L_p(bcmf2Rem.synthetic(4000,1))
Lbcmf3Rem = L_p(bcmf3Rem.synthetic(4000,1))
Lbcmf4Rem = L_p(bcmf4Rem.synthetic(4000,1))
Lbcmf1Full = L_p(bcmf1Full.synthetic(4000,1))
Lbcmf2Full = L_p(bcmf2Full.synthetic(4000,1))
Lbcmf3Full = L_p(bcmf3Full.synthetic(4000,1))
Lbcmf4Full = L_p(bcmf4Full.synthetic(4000,1))

##==============================================================================
## There are various variations to calculate the psf: Need to be checked individually
## #psfSingle = PointSpreadFunction(grid=g, mpos=m, calcmode='single')
## #LPsfSingle = L_p(psfSingle.psf[:])
## 
## #psfBlock = PointSpreadFunction(grid=g, mpos=m, calcmode='block')
## #LPsfBlock = L_p(psfBlock.psf[:])
## 
## #psfFull = PointSpreadFunction(grid=g, mpos=m, calcmode='full')
## #LPsfFull = L_p(psfFull.psf[:])
##
## #all_bf = (LPsfFull,)
##==============================================================================
psf1 = PointSpreadFunction(grid=g, mpos=m, c=346.04, steer='classic')
psf2 = PointSpreadFunction(grid=g, mpos=m, c=346.04, steer='inverse')
psf3 = PointSpreadFunction(grid=g, mpos=m, c=346.04, steer='true level')
psf4 = PointSpreadFunction(grid=g, mpos=m, c=346.04, steer='true location')
Lpsf1 = L_p(psf1.psf[:])
Lpsf2 = L_p(psf2.psf[:])
Lpsf3 = L_p(psf3.psf[:])
Lpsf4 = L_p(psf4.psf[:])

bcpsf1Rem = BeamformerClean(beamformer=bb1Rem)
bcpsf2Rem = BeamformerClean(beamformer=bb2Rem)
bcpsf3Rem = BeamformerClean(beamformer=bb3Rem)
bcpsf4Rem = BeamformerClean(beamformer=bb4Rem)
bcpsf1Full = BeamformerClean(beamformer=bb1Full)
bcpsf2Full = BeamformerClean(beamformer=bb2Full)
bcpsf3Full = BeamformerClean(beamformer=bb3Full)
bcpsf4Full = BeamformerClean(beamformer=bb4Full)
Lbcpsf1Rem = L_p(bcpsf1Rem.synthetic(4000,1))
Lbcpsf2Rem = L_p(bcpsf2Rem.synthetic(4000,1))
Lbcpsf3Rem = L_p(bcpsf3Rem.synthetic(4000,1))
Lbcpsf4Rem = L_p(bcpsf4Rem.synthetic(4000,1))
Lbcpsf1Full = L_p(bcpsf1Full.synthetic(4000,1))
Lbcpsf2Full = L_p(bcpsf2Full.synthetic(4000,1))
Lbcpsf3Full = L_p(bcpsf3Full.synthetic(4000,1))
Lbcpsf4Full = L_p(bcpsf4Full.synthetic(4000,1))

bd1Rem = BeamformerDamas(beamformer=bb1Rem, n_iter=100)
bd2Rem = BeamformerDamas(beamformer=bb2Rem, n_iter=100)
bd3Rem = BeamformerDamas(beamformer=bb3Rem, n_iter=100)
bd4Rem = BeamformerDamas(beamformer=bb4Rem, n_iter=100)
bd1Full = BeamformerDamas(beamformer=bb1Full, n_iter=100)
bd2Full = BeamformerDamas(beamformer=bb2Full, n_iter=100)
bd3Full = BeamformerDamas(beamformer=bb3Full, n_iter=100)
bd4Full = BeamformerDamas(beamformer=bb4Full, n_iter=100)
Lbd1Rem = L_p(bd1Rem.synthetic(4000,1))
Lbd2Rem = L_p(bd2Rem.synthetic(4000,1))
Lbd3Rem = L_p(bd3Rem.synthetic(4000,1))
Lbd4Rem = L_p(bd4Rem.synthetic(4000,1))
Lbd1Full = L_p(bd1Full.synthetic(4000,1))
Lbd2Full = L_p(bd2Full.synthetic(4000,1))
Lbd3Full = L_p(bd3Full.synthetic(4000,1))
Lbd4Full = L_p(bd4Full.synthetic(4000,1))



all_bf = (Lbb1Rem, Lbb2Rem, Lbb3Rem, Lbb4Rem, Lbb1Full, Lbb2Full, Lbb3Full, Lbb4Full,
          Lbf1Rem, Lbf2Rem, Lbf3Rem, Lbf4Rem, Lbf1Full, Lbf2Full, Lbf3Full, Lbf4Full,
          Lbca1Full, Lbca2Full, Lbca3Full, Lbca4Full,
          Lbe1Rem, Lbe2Rem, Lbe3Rem, Lbe4Rem, Lbe1Full, Lbe2Full, Lbe3Full, Lbe4Full,
          Lbm1Full, Lbm2Full, Lbm3Full, Lbm4Full,
          Lbcsc1Rem, Lbcsc2Rem, Lbcsc3Rem, Lbcsc4Rem, Lbcsc1Full, Lbcsc2Full, Lbcsc3Full, Lbcsc4Full,
          Lbort1Rem, Lbort2Rem, Lbort3Rem, Lbort4Rem, Lbort1Full, Lbort2Full, Lbort3Full, Lbort4Full,
          Lbcmf1Rem, Lbcmf2Rem, Lbcmf3Rem, Lbcmf4Rem, Lbcmf1Full, Lbcmf2Full, Lbcmf3Full, Lbcmf4Full, 
          Lpsf1, Lpsf2, Lpsf3, Lpsf4,
          Lbcpsf1Rem, Lbcpsf2Rem, Lbcpsf3Rem, Lbcpsf4Rem, Lbcpsf1Full, Lbcpsf2Full, Lbcpsf3Full, Lbcpsf4Full,
          Lbd1Rem, Lbd2Rem, Lbd3Rem, Lbd4Rem, Lbd1Full, Lbd2Full, Lbd3Full, Lbd4Full)

fi = open('all_bfWeave.sav','w')  # This file saves the outputs of the current acoular version
#fi = open('all_bfNumba.sav','w')  # This file saves the outputs of the new acoular version, which has to be validated

dump(all_bf,fi,-1) # uses newest pickle protocol -1 (default = 0)
fi.close()
#"""

#%% Compare Weave vs Numba
fi = open('all_bfWeave.sav','r')
all_bfWeave = load(fi)
fi.close()

fi = open('all_bfNumba.sav','r')
all_bfNumba = load(fi)
fi.close()

# remove all negative levels
err = []  # keep in mind that these are levels!!!
for cnt in range(len(all_bfNumba)):
    all_bfNumba[cnt][all_bfNumba[cnt] < 0] = all_bfWeave[cnt][all_bfWeave[cnt] < 0] = 1e-20
    relDiff = (all_bfWeave[cnt] - all_bfNumba[cnt]) / (all_bfWeave[cnt] + all_bfNumba[cnt]) * 2
    err.append(amax(amax(amax(abs(relDiff), 0), 0), 0))



# -*- coding: utf-8 -*-
#pylint: disable-msg=E0611, E1101, C0103, R0901, R0902, R0903, R0904, W0232
#------------------------------------------------------------------------------
# Copyright (c) Acoular Development Team.
#------------------------------------------------------------------------------
""" Input from soundcard hardware using the SoundDevice library

.. autosummary::
    :toctree: generated/

    SoundDeviceSamplesGenerator
"""
import sounddevice as sd
from traits.api import Int, Long, Bool, Property, Any, observe, cached_property
from .tprocess import SamplesGenerator
from .internal import digest

class SoundDeviceSamplesGenerator(SamplesGenerator):

    """
    Controller for sound card hardware using sounddevice library

    Uses the device with index :attr:`device` to read samples
    from input stream, generates output stream via the generator
    :meth:`result`.
    """

    #: input device index, refers to sounddevice list
    device = Int(0, desc="input device index")

    #: Number of input channels, maximum depends on device
    numchannels = Long(1,
       desc="number of analog input channels that collects data")

    #: Number of samples to collect; defaults to -1.
    # If is set to -1 device collects till user breaks streaming by setting Trait: collectsamples = False
    numsamples = Long(-1,
        desc="number of samples to collect")    

    #: Indicates if samples are collected, helper trait to break result loop
    collectsamples = Bool(True,
        desc="Indicates if samples are collected")

    #: Sampling frequency of the signal, changes with sinusdevices
    sample_freq = Property(
        desc="sampling frequency")

    #: Indicates that the sounddevice buffer has overflown
    overflow = Bool(False,
        desc="Indicates if sounddevice buffer overflow")

    #: Indicates that the stream is collecting samples
    running = Bool(False,
        desc="Indicates that the stream is collecting samples")

    #: The sounddevice InputStream object for inspection
    stream = Any

    # internal identifier
    digest = Property( depends_on=['device', 'numchannels', 'numsamples'])
    
    @cached_property
    def _get_digest( self ):
        return digest(self)

    # checks that numchannels are not more than device can provide
    @observe('device,numchannels')
    def _get_numchannels( self, change ):
        self.numchannels = min(self.numchannels,sd.query_devices(self.device)['max_input_channels'])

    def _get_sample_freq( self ):
        return sd.query_devices(self.device)['default_samplerate']

    def device_properties( self ):
        """
        Returns
        -------
        Dictionary of device properties according to sounddevice
        """
        return sd.query_devices(self.device)

    def result(self, num):
        """
        Python generator that yields the output block-wise. Use at least a 
        block-size of one ring cache block. 
        
        Parameters
        ----------
        num : integer
            This parameter defines the size of the blocks to be yielded
            (i.e. the number of samples per block).
        
        Returns
        -------
        Samples in blocks of shape (num, :attr:`numchannels`). 
            The last block may be shorter than num.
        """
        print(self.device_properties(),self.sample_freq)
        self.stream = stream_obj = sd.InputStream(
                device=self.device, channels=self.numchannels, 
                clip_off=True, samplerate=self.sample_freq)

        with stream_obj as stream:
            self.running = True
            if self.numsamples == -1: 
                while self.collectsamples: # yield data as long as collectsamples is True
                    data,self.overflow = stream.read(num)
                    yield data[:num]   
                                           
            elif self.numsamples > 0: # amount of samples to collect is specified by user            
                samples_count = 0 # numsamples counter 
                while samples_count < self.numsamples: 
                    anz = min(num, self.numsamples-samples_count)
                    data, self.overflow = stream.read(num)
                    yield data[:anz]   
                    samples_count += anz
        self.running = False
        return  



# coding=UTF-8
#------------------------------------------------------------------------------
# Copyright (c) Acoular Development Team.
#------------------------------------------------------------------------------

from hashlib import md5

def digest( obj, name='digest'):
    str_ = [str(obj.__class__).encode("UTF-8")]
    for do_ in obj.trait(name).depends_on:
        vobj = obj
        try:
            for i in do_.split('.'):               
                vobj = list(vobj.trait_get(i.rstrip('[]')).values())[0]
            str_.append(str(vobj).encode("UTF-8"))
        except:
            pass
    return '_' + md5(''.encode("UTF-8").join(str_)).hexdigest()

def ldigest(l):
    str_ = []
    for i in l:
        str_.append(str(i.digest).encode('UTF-8'))
    return '_' + md5(''.encode("UTF-8").join(str_)).hexdigest()





# -*- coding: utf-8 -*-
#------------------------------------------------------------------------------
# Copyright (c) Acoular Development Team.
#------------------------------------------------------------------------------
from os.path import join
import numpy as np
#acoular imports
import acoular
acoular.config.global_caching = 'none' # to make sure that nothing is cached

from acoular import Calib, MicGeom, PowerSpectra, RectGrid, BeamformerBase,\
SteeringVector, Environment, MaskedTimeSamples

#load exampledata
datafile = join('..','example_data.h5')
micgeofile = join( '..','..','acoular','xml','array_56.xml')

# values from example 1
t1 = MaskedTimeSamples(name=datafile)
t1.start = 0 # first sample, default
t1.stop = 16000 # last valid sample = 15999
m = MicGeom(from_file=micgeofile)
g = RectGrid(x_min=-0.6, x_max=-0.0, y_min=-0.3, y_max=0.3, z=0.68,
             increment=0.01 )
env = Environment(c=346.04)
st = SteeringVector(grid=g, mics=m, env=env, steer_type='classic')
f = PowerSpectra(time_data=t1, 
               window='Hanning', overlap='50%', block_size=4096, #FFT-parameters
               cached = False )  
b = BeamformerBase(freq_data=f, steer=st, r_diag=True, cached = False)

csm = np.array(b.freq_data.csm[20], dtype='complex128')
param_steer_type, steer_vector = b._beamformer_params()
sln = b.sig_loss_norm()
svt = steer_vector(8000.)
eva = b.freq_data.eva[20]
eve = b.freq_data.eve[20]

def test_beamformerFreq():
    acoular.fastFuncs.beamformerFreq(param_steer_type, False, sln, svt, csm)

def test_beamformerFreqE():
    acoular.fastFuncs.beamformerFreq(param_steer_type, False, sln, svt, (eva[30:],eve[:,30:]))

def test_csm():
    t1.stop=-1
    f.calc_csm()
